import unittest

import mock
import numpy as np

from smqtk.algorithms.nn_index.hash_index.sklearn_balltree import \
    SkLearnBallTreeHashIndex
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.utils.bit_utils import int_to_bit_vector_large


class TestBallTreeHashIndex (unittest.TestCase):

    def test_is_usable(self):
        # Should always be true because major dependency (sklearn) is a package
        # requirement.
        self.assertTrue(SkLearnBallTreeHashIndex.is_usable())

    def test_default_configuration(self):
        c = SkLearnBallTreeHashIndex.get_default_config()
        self.assertEqual(len(c), 3)
        self.assertIsInstance(c['cache_element'], dict)
        self.assertIsNone(c['cache_element']['type'])
        self.assertEqual(c['leaf_size'], 40)
        self.assertIsNone(c['random_seed'])

    def test_init_without_cache(self):
        i = SkLearnBallTreeHashIndex(cache_element=None, leaf_size=52,
                                     random_seed=42)
        self.assertIsNone(i.cache_element)
        self.assertEqual(i.leaf_size, 52)
        self.assertEqual(i.random_seed, 42)
        self.assertIsNone(i.bt)

    def test_init_with_empty_cache(self):
        empty_cache = DataMemoryElement()
        i = SkLearnBallTreeHashIndex(cache_element=empty_cache,
                                     leaf_size=52,
                                     random_seed=42)
        self.assertEqual(i.cache_element, empty_cache)
        self.assertEqual(i.leaf_size, 52)
        self.assertEqual(i.random_seed, 42)
        self.assertIsNone(i.bt)

    def test_get_config(self):
        bt = SkLearnBallTreeHashIndex()
        bt_c = bt.get_config()

        self.assertEqual(len(bt_c), 3)
        self.assertIn('cache_element', bt_c)
        self.assertIn('leaf_size', bt_c)
        self.assertIn('random_seed', bt_c)

        self.assertIsInstance(bt_c['cache_element'], dict)
        self.assertIsNone(bt_c['cache_element']['type'])

    def test_init_consistency(self):
        # Test that constructing an instance with a configuration yields the
        # same config via ``get_config``.

        # - Default config should be a valid configuration for this impl.
        c = SkLearnBallTreeHashIndex.get_default_config()
        self.assertEqual(
            SkLearnBallTreeHashIndex.from_config(c).get_config(),
            c
        )
        # With non-null cache element
        c['cache_element']['type'] = 'DataMemoryElement'
        self.assertEqual(
            SkLearnBallTreeHashIndex.from_config(c).get_config(),
            c
        )

    def test_build_index_no_input(self):
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        self.assertRaises(
            ValueError,
            bt.build_index, []
        )

    def test_build_index(self):
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        # Make 1000 random bit vectors of length 256
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)
        bt.build_index(m)

        # deterministically sort index of built and source data to determine
        # that an index was built.
        self.assertIsNotNone(bt.bt)
        np.testing.assert_array_almost_equal(
            sorted(np.array(bt.bt.data).tolist()),
            sorted(m.tolist())
        )

    def test_update_index_no_input(self):
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        self.assertRaises(
            ValueError,
            bt.update_index, []
        )

    def test_update_index_new_index(self):
        # Virtually the same as `test_build_index` but using update_index.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        # Make 1000 random bit vectors of length 256
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256).astype(bool)
        bt.update_index(m)

        # deterministically sort index of built and source data to determine
        # that an index was built.
        self.assertIsNotNone(bt.bt)
        np.testing.assert_array_almost_equal(
            sorted(np.array(bt.bt.data).tolist()),
            sorted(m.tolist())
        )

    def test_update_index_additive(self):
        # Test updating an existing index, i.e. rebuilding using the union of
        # previous and new data.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        # Make 1000 random bit vectors of length 256
        m1 = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)\
               .astype(bool)
        m2 = np.random.randint(0, 2, 100 * 256).reshape(100, 256).astype(bool)

        # Build initial index
        bt.build_index(m1)
        # Current model should only contain m1's data.
        np.testing.assert_array_almost_equal(
            sorted(np.array(bt.bt.data).tolist()),
            sorted(m1.tolist())
        )

        # "Update" index with new hashes
        bt.update_index(m2)
        # New model should contain the union of the data.
        np.testing.assert_array_almost_equal(
            sorted(np.array(bt.bt.data).tolist()),
            sorted(np.concatenate([m1, m2], 0).tolist())
        )

    def test_remove_from_index_no_index(self):
        # A key error should be raised if there is no ball-tree index yet.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        rm_hash = np.random.randint(0, 2, 256)
        self.assertRaisesRegexp(
            KeyError,
            str(rm_hash[0]),
            bt.remove_from_index,
            [rm_hash]
        )

    def test_remove_from_index_invalid_key_single(self):
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        index = np.ndarray((1000, 256), bool)
        for i in range(1000):
            index[i] = int_to_bit_vector_large(i, 256)
        bt.build_index(index)
        # Copy post-build index for checking no removal occurred
        bt_data = np.copy(bt.bt.data)

        self.assertRaises(
            KeyError,
            bt.remove_from_index, [
                int_to_bit_vector_large(1001, 256),
            ]
        )
        np.testing.assert_array_equal(
            bt_data,
            np.asarray(bt.bt.data)
        )

    def test_remove_from_index_invalid_key_multiple(self):
        # Test that mixed valid and invalid keys raises KeyError and does not
        # modify the index.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        index = np.ndarray((1000, 256), bool)
        for i in range(1000):
            index[i] = int_to_bit_vector_large(i, 256)
        bt.build_index(index)
        # Copy post-build index for checking no removal occurred
        bt_data = np.copy(bt.bt.data)

        self.assertRaises(
            KeyError,
            bt.remove_from_index, [
                int_to_bit_vector_large(42, 256),
                int_to_bit_vector_large(1008, 256),
            ]
        )
        np.testing.assert_array_equal(
            bt_data,
            np.asarray(bt.bt.data)
        )

    def test_remove_from_index(self):
        # Test that we actually remove from the index.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        index = np.ndarray((1000, 256), bool)
        for i in range(1000):
            index[i] = int_to_bit_vector_large(i, 256)
        bt.build_index(index)
        # Copy post-build index for checking no removal occurred
        bt_data = np.copy(bt.bt.data)

        bt.remove_from_index([
            int_to_bit_vector_large(42, 256),
            int_to_bit_vector_large(998, 256),
        ])
        # Make sure expected arrays are missing from data block.
        new_data = np.asarray(bt.bt.data)
        self.assertEqual(new_data.shape, (998, 256))
        new_data_set = set(tuple(r) for r in new_data.tolist())
        self.assertNotIn(tuple(int_to_bit_vector_large(42, 256)),
                         new_data_set)
        self.assertNotIn(tuple(int_to_bit_vector_large(998, 256)),
                         new_data_set)

    def test_remove_from_index_last_element(self):
        """
        Test removing the final the only element / final elements from the
        index.
        """
        # Add one hash, remove one hash.
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        index = np.ndarray((1, 256), bool)
        index[0] = int_to_bit_vector_large(1, 256)
        bt.build_index(index)
        self.assertEqual(bt.count(), 1)
        bt.remove_from_index(index)
        self.assertEqual(bt.count(), 0)
        self.assertIsNone(bt.bt)

        # Add many hashes, remove many hashes in batches until zero
        bt = SkLearnBallTreeHashIndex(random_seed=0)
        index = np.ndarray((1000, 256), bool)
        for i in range(1000):
            index[i] = int_to_bit_vector_large(i, 256)
        bt.build_index(index)
        # Remove first 250
        bt.remove_from_index(index[:250])
        self.assertEqual(bt.count(), 750)
        self.assertIsNotNone(bt.bt)
        # Remove second 250
        bt.remove_from_index(index[250:500])
        self.assertEqual(bt.count(), 500)
        self.assertIsNotNone(bt.bt)
        # Remove third 250
        bt.remove_from_index(index[500:750])
        self.assertEqual(bt.count(), 250)
        self.assertIsNotNone(bt.bt)
        # Remove final 250
        bt.remove_from_index(index[750:])
        self.assertEqual(bt.count(), 0)
        self.assertIsNone(bt.bt)

    def test_remove_from_index_last_element_with_cache(self):
        """
        Test removing final element also clears the cache element.
        """
        c = DataMemoryElement()
        bt = SkLearnBallTreeHashIndex(cache_element=c, random_seed=0)
        index = np.ndarray((1, 256), bool)
        index[0] = int_to_bit_vector_large(1, 256)

        bt.build_index(index)
        self.assertEqual(bt.count(), 1)
        self.assertFalse(c.is_empty())

        bt.remove_from_index(index)
        self.assertEqual(bt.count(), 0)
        self.assertTrue(c.is_empty())

    def test_count_empty(self):
        bt = SkLearnBallTreeHashIndex()
        self.assertEqual(bt.count(), 0)

    def test_count_nonempty(self):
        bt = SkLearnBallTreeHashIndex()
        # Make 1000 random bit vectors of length 256
        m = np.random.randint(0, 2, 234 * 256).reshape(234, 256)
        bt.build_index(m)

        self.assertEqual(bt.count(), 234)

    def test_nn_no_index(self):
        i = SkLearnBallTreeHashIndex()

        self.assertRaisesRegexp(
            ValueError,
            "No index currently set to query from",
            i.nn, [0, 0, 0]
        )

    @mock.patch('smqtk.algorithms.nn_index.hash_index.sklearn_balltree.np'
                '.savez')
    def test_save_model_no_cache(self, m_savez):
        bt = SkLearnBallTreeHashIndex()
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)
        bt._build_bt_internal(m)
        # Underlying serialization function should not have been called
        # because no cache element set.
        self.assertFalse(m_savez.called)

    def test_save_model_with_readonly_cache(self):
        cache_element = DataMemoryElement(readonly=True)
        bt = SkLearnBallTreeHashIndex(cache_element)
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)
        self.assertRaises(
            ValueError,
            bt._build_bt_internal, m
        )

    @mock.patch('smqtk.algorithms.nn_index.hash_index.sklearn_balltree.np'
                '.savez')
    def test_save_model_with_cache(self, m_savez):
        cache_element = DataMemoryElement()
        bt = SkLearnBallTreeHashIndex(cache_element, random_seed=0)
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)
        bt._build_bt_internal(m)
        self.assertTrue(m_savez.called)
        self.assertEqual(m_savez.call_count, 1)

    def test_load_model(self):
        # Create two index instances, building model with one, and loading
        # the other with the cache of the first instance. Each should have
        # distinct model instances, but should otherwise have equal model
        # values and parameters.
        cache_element = DataMemoryElement()
        bt1 = SkLearnBallTreeHashIndex(cache_element, random_seed=0)
        m = np.random.randint(0, 2, 1000 * 256).reshape(1000, 256)
        bt1.build_index(m)

        bt2 = SkLearnBallTreeHashIndex(cache_element)
        self.assertIsNotNone(bt2.bt)

        q = np.random.randint(0, 2, 256).astype(bool)
        bt_neighbors, bt_dists = bt1.nn(q, 10)
        bt2_neighbors, bt2_dists = bt2.nn(q, 10)

        self.assertIsNot(bt1, bt2)
        self.assertIsNot(bt1.bt, bt2.bt)
        np.testing.assert_equal(bt2_neighbors, bt_neighbors)
        np.testing.assert_equal(bt2_dists, bt_dists)



"""
Script for building and saving the model for the ``SkLearnBallTreeHashIndex``
implementation of ``HashIndex``.
"""
import logging

from smqtk.algorithms.nn_index.hash_index.sklearn_balltree import \
    SkLearnBallTreeHashIndex
import smqtk.representation
import smqtk.utils.bin_utils
import smqtk.utils.bit_utils
import smqtk.utils.plugin


def default_config():
    return {
        "hash2uuid_kv_store": smqtk.utils.plugin.make_config(
            smqtk.representation.get_key_value_store_impls()
        ),
        "sklearn_balltree": SkLearnBallTreeHashIndex.get_default_config(),
        "itq_bit_length": 256,
    }


def cli_parser():
    parser = smqtk.utils.bin_utils.basic_cli_parser(__doc__)
    return parser


def main():
    args = cli_parser().parse_args()
    config = smqtk.utils.bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    # Loading from configurations
    bit_len = int(config['itq_bit_length'])
    log.info("Loading hash2uuid KeyValue store")
    #: :type: smqtk.representation.KeyValueStore
    hash2uuid_kv_store = smqtk.utils.plugin.from_plugin_config(
        config['hash2uuid_kv_store'],
        smqtk.representation.get_key_value_store_impls()
    )
    log.info("Initializing ball tree")
    btree = SkLearnBallTreeHashIndex.from_config(config['sklearn_balltree'])

    log.info("Computing hash-code vectors")
    hash_vectors = []
    reporter = smqtk.utils.bin_utils.ProgressReporter(log.debug, 1.0)
    reporter.start()
    for h in hash2uuid_kv_store.keys():
        hash_vectors.append(
            smqtk.utils.bit_utils.int_to_bit_vector_large(h, bit_len))
        reporter.increment_report()
    reporter.report()

    log.info("Building ball tree index")
    btree.build_index(hash_vectors)


if __name__ == '__main__':
    main()



import abc
import collections
import os

from smqtk.representation import SmqtkRepresentation
from smqtk.utils import plugin


class DataSet (collections.Set, SmqtkRepresentation, plugin.Pluggable):
    """
    Abstract interface for data sets, that contain an arbitrary number of
    ``DataElement`` instances of arbitrary implementation type, keyed on
    ``DataElement`` UUID values.

    This should only be used with DataElements whose byte content is expected
    not to change. If they do, then UUID keys may no longer represent the
    elements associated with them.

    """

    def __len__(self):
        """
        :return: Number of elements in this DataSet.
        :rtype: int
        """
        return self.count()

    def __getitem__(self, uuid):
        return self.get_data(uuid)

    def __contains__(self, d):
        """
        Different than has_uuid() because this takes another DataElement
        instance, not a UUID.

        :param d: DataElement to test for containment
        :type d: smqtk.representation.DataElement

        :return: True of this DataSet contains the given data element. Since,
        :rtype: bool

        """
        return self.has_uuid(d.uuid())

    @abc.abstractmethod
    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """

    @abc.abstractmethod
    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """

    @abc.abstractmethod
    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """

    @abc.abstractmethod
    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the
            type that the set implementation expects or cares about.
        :type uuid: collections.Hashable

        :return: True if the given uuid matches an element in this set, or
            False if it does not.
        :rtype: bool

        """

    @abc.abstractmethod
    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        *NOTE: Implementing methods should check that input elements are in
        fact DataElement instances.*

        :param elems: Data element(s) to add
        :type elems: smqtk.representation.DataElement

        """

    @abc.abstractmethod
    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.
        :type uuid: collections.Hashable

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement

        """


def get_data_set_impls(reload_modules=False):
    """
    Discover and return discovered ``DataSet`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable ``DATA_SET_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``DATA_SET_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``DataSet``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "DATA_SET_PATH"
    helper_var = "DATA_SET_CLASS"
    return plugin.get_plugins(__name__, this_dir, env_var, helper_var, DataSet,
                              reload_modules=reload_modules)



"""
Mongo Session wrapper

Allows session updating within AJAX routines.

Taken from: http://flask.pocoo.org/snippets/110/

"""

from uuid import uuid4
from datetime import datetime, timedelta

from flask.sessions import SessionInterface, SessionMixin
from werkzeug.datastructures import CallbackDict
from pymongo import MongoClient


class MongoSession(CallbackDict, SessionMixin):

    def __init__(self, initial=None, sid=None):
        CallbackDict.__init__(self, initial)
        self.sid = sid
        self.modified = False


class MongoSessionInterface(SessionInterface):

    def __init__(self, host='localhost', port=27017,
                 db='', collection='sessions'):
        client = MongoClient(host, port)
        self.store = client[db][collection]

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if sid:
            stored_session = self.store.find_one({'sid': sid})
            if stored_session:
                if stored_session.get('expiration') > datetime.utcnow():
                    return MongoSession(initial=stored_session['data'],
                                        sid=stored_session['sid'])
        sid = str(uuid4())
        return MongoSession(sid=sid)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        if not session:
            response.delete_cookie(app.session_cookie_name, domain=domain)
            return
        if self.get_expiration_time(app, session):
            expiration = self.get_expiration_time(app, session)
        else:
            expiration = datetime.utcnow() + timedelta(hours=1)
        self.store.update({'sid': session.sid},
                          {'sid': session.sid,
                           'data': session,
                           'expiration': expiration}, True)
        response.set_cookie(app.session_cookie_name, session.sid,
                            expires=self.get_expiration_time(app, session),
                            httponly=True, domain=domain)



from __future__ import division, print_function
import random
import unittest

import numpy as np
from six.moves import range

from six.moves import range
from smqtk.utils import metrics as df


def gen(n):
    return random.randint(0, 2**n-1)


class TestHistogramIntersectionDistance (unittest.TestCase):

    v1 = np.array([0, 0])
    v2 = np.array([1, 0])
    v3 = np.array([0, 1])
    v4 = np.array([.5, .5])

    m1 = np.array([v2, v3, v4])
    m2 = np.array([v2, v4])

    hi_methods = [
        df.histogram_intersection_distance_fast,
        df.histogram_intersection_distance,
    ]

    def test_hi_result_zerovector(self):
        # HI distance of anything with the zero vector is 1.0 since a valid
        # histogram does not intersect with nothing.
        # REMEMBER we're talking about distance here, not similarity
        for m in self.hi_methods:
            print("Tests for method: %s" % m)
            self.assertEqual(m(self.v1, self.v1), 1.)

            self.assertEqual(m(self.v1, self.v2), 1.)
            self.assertEqual(m(self.v2, self.v1), 1.)

            self.assertEqual(m(self.v1, self.v4), 1.)
            self.assertEqual(m(self.v4, self.v1), 1.)

    def test_hi_result_normal(self):
        for m in self.hi_methods:
            print("Tests for method: %s" % m)

            self.assertEqual(m(self.v2, self.v3), 1.)
            self.assertEqual(m(self.v3, self.v2), 1.)

            self.assertEqual(m(self.v2, self.v4), 0.5)
            self.assertEqual(m(self.v4, self.v2), 0.5)

            self.assertEqual(m(self.v3, self.v4), 0.5)
            self.assertEqual(m(self.v4, self.v3), 0.5)

            self.assertEqual(m(self.v4, self.v4), 0.0)

    def test_hi_input_format(self):
        # the general form method should be able to take any combination of
        # vectors and matrices, following documented rules.

        self.assertEqual(
            df.histogram_intersection_distance(self.v4, self.v3),
            0.5
        )

        np.testing.assert_array_equal(
            df.histogram_intersection_distance(self.v2, self.m1),
            [0., 1., 0.5]
        )
        np.testing.assert_array_equal(
            df.histogram_intersection_distance(self.m1, self.v2),
            [0., 1., 0.5]
        )

        np.testing.assert_array_equal(
            df.histogram_intersection_distance(self.m1, self.m1),
            [0, 0, 0]
        )

        self.assertRaises(ValueError, df.histogram_intersection_distance,
                          self.m1, self.m2)


class TestHammingDistance (unittest.TestCase):

    def test_hd_0(self):
        self.assertEqual(df.hamming_distance(0, 0), 0)

    def test_rand(self):
        n = 64
        for i in range(1000):
            a = gen(n)
            b = gen(n)
            actual = bin(a ^ b).count('1')
            self.assertEqual(df.hamming_distance(a, b), actual)

    def test_rand_large(self):
        n = 1024
        for i in range(1000):
            a = gen(n)
            b = gen(n)
            actual = bin(a ^ b).count('1')
            self.assertEqual(df.hamming_distance(a, b), actual)



import abc
import numpy
import os

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.representation import DescriptorElementFactory
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.utils.parallel import parallel_map
from smqtk.utils.plugin import get_plugins


DFLT_DESCRIPTOR_FACTORY = DescriptorElementFactory(DescriptorMemoryElement, {})


class DescriptorGenerator (SmqtkAlgorithm):
    """
    Base abstract Feature Descriptor interface
    """

    @abc.abstractmethod
    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """

    def compute_descriptor(self, data, descr_factory=DFLT_DESCRIPTOR_FACTORY,
                           overwrite=False):
        """
        Given some data, return a descriptor element containing a descriptor
        vector.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor.

        :param data: Some kind of input data for the feature descriptor.
        :type data: smqtk.representation.DataElement

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param ot staverwrite: Whether or not to force re-computation of a descriptor
            vector for the given data even when there exists a precomputed
            vector in the generated DescriptorElement as generated from the
            provided factory. This will overwrite the persistently stored vector
            if the provided factory produces a DescriptorElement implementation
            with such storage.
        :type overwrite: bool

        :return: Result descriptor element. UUID of this output descriptor is
            the same as the UUID of the input data element.
        :rtype: smqtk.representation.DescriptorElement

        """
        # Check content type against listed valid types
        ct = data.content_type()
        if ct not in self.valid_content_types():
            self._log.error("Cannot compute descriptor from content type '%s' "
                            "data: %s)" % (ct, data))
            raise ValueError("Cannot compute descriptor from content type '%s' "
                             "data: %s)" % (ct, data))

        # Produce the descriptor element container via the provided factory
        # - If the generated element already contains a vector, because the
        #   implementation provides some kind of persistent caching mechanism
        #   or something, don't compute another descriptor vector unless the
        #   overwrite flag is True
        descr_elem = descr_factory.new_descriptor(self.name, data.uuid())
        if overwrite or not descr_elem.has_vector():
            vec = self._compute_descriptor(data)
            descr_elem.set_vector(vec)
        else:
            self._log.debug("Found existing vector in generated element: %s",
                            descr_elem)

        return descr_elem

    def compute_descriptor_async(self, data_iter,
                                 descr_factory=DFLT_DESCRIPTOR_FACTORY,
                                 overwrite=False, procs=None, **kwds):
        """
        Asynchronously compute feature data for multiple data items.

        Base implementation additional keyword arguments:
            use_mp [= False]
                If multi-processing should be used vs. multi-threading.

        :param data_iter: Iterable of data elements to compute features for.
            These must have UIDs assigned for feature association in return
            value.
        :type data_iter: collections.Iterable[smqtk.representation.DataElement]

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param overwrite: Whether or not to force re-computation of a descriptor
            vectors for the given data even when there exists precomputed
            vectors in the generated DescriptorElements as generated from the
            provided factory. This will overwrite the persistently stored
            vectors if the provided factory produces a DescriptorElement
            implementation such storage.
        :type overwrite: bool

        :param procs: Optional specification of how many processors to use
            when pooling sub-tasks. If None, we attempt to use all available
            cores.
        :type procs: int | None

        :raises ValueError: An input DataElement was of a content type that we
            cannot handle.

        :return: Mapping of input DataElement UUIDs to the computed descriptor
            element for that data. DescriptorElement UUID's are congruent with
            the UUID of the data element it is the descriptor of.
        :rtype: dict[collections.Hashable,
                     smqtk.representation.DescriptorElement]

        """
        self._log.info("Async compute features")

        use_mp = kwds.get('use_mp', False)

        def work(d):
            return d, self.compute_descriptor(d, descr_factory, overwrite)

        results = parallel_map(work, data_iter, cores=procs, ordered=False,
                               use_multiprocessing=use_mp)

        de_map = {}
        for data, descriptor in results:
            de_map[data.uuid()] = descriptor

        return de_map

    @abc.abstractmethod
    def _compute_descriptor(self, data):
        """
        Internal method that defines the generation of the descriptor vector for
        a given data element. This returns a numpy array.

        This method is only called if the data element has been verified to be
        of a valid content type for this descriptor implementation.

        :raises RuntimeError: Feature extraction failure of some kind.

        :param data: Some kind of input data for the feature descriptor.
        :type data: smqtk.representation.DataElement

        :return: Feature vector.
        :rtype: numpy.core.multiarray.ndarray

        """


def get_descriptor_generator_impls(reload_modules=False):
    """
    Discover and return discovered ``DescriptorGenerator`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable
          ``DESCRIPTOR_GENERATOR_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``DESCRIPTOR_GENERATOR_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``DescriptorGenerator``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "DESCRIPTOR_GENERATOR_PATH"
    helper_var = "DESCRIPTOR_GENERATOR_CLASS"
    return get_plugins(__name__, this_dir, env_var, helper_var,
                       DescriptorGenerator, reload_modules=reload_modules)



from __future__ import division, print_function
import random
import unittest

from six.moves import map, range

from smqtk.utils.parallel import parallel_map


class TestParallelMap (unittest.TestCase):

    @classmethod
    def setup_class(cls):
        n = 10000

        # Random characters in range [a, z]
        cls.test_string = [chr(random.randint(97, 122)) for _ in range(n)]
        cls.test_func = ord
        # Since this parallel function is intended to perform similar to the
        # built-in map function.
        cls.expected = list(map(cls.test_func, cls.test_string))

    def test_simple_ordered_threaded(self):
        # Make sure results are still in order as requested
        r = list(parallel_map(self.test_func, self.test_string,
                              ordered=True, use_multiprocessing=False))
        self.assertEqual(r, self.expected)

    def test_simple_ordered_multiprocess(self):
        r = list(parallel_map(self.test_func, self.test_string,
                              ordered=True, use_multiprocessing=True))
        self.assertEqual(r, self.expected)

    def test_simple_unordered_threaded(self):
        r = list(parallel_map(self.test_func, self.test_string,
                              ordered=False, use_multiprocessing=False))
        self.assertEqual(set(r), set(self.expected))

    def test_simple_unordered_multiprocess(self):
        r = list(parallel_map(self.test_func, self.test_string,
                              ordered=False, use_multiprocessing=True))
        self.assertEqual(set(r), set(self.expected))

    def test_exception_handing_threaded(self):
        def raise_ex(_):
            raise RuntimeError("Expected exception")

        self.assertRaises(
            RuntimeError,
            list,
            parallel_map(raise_ex, [1], use_multiprocessing=False)
        )

    def test_exception_handing_multiprocess(self):
        def raise_ex(_):
            raise RuntimeError("Expected exception")

        self.assertRaises(
            RuntimeError,
            list,
            parallel_map(raise_ex, [1], use_multiprocessing=True)
        )

    def test_multisequence(self):
        def test_func(a, b, c):
            return a + b + c

        s1 = [1] * 10
        s2 = [2] * 10
        s3 = [3] * 10
        r = list(parallel_map(test_func, s1, s2, s3,
                              use_multiprocessing=False))

        expected = [6] * 10
        self.assertEqual(r, expected)

    def test_multisequence_short_cutoff(self):
        def test_func(a, b, c):
            return a + b + c

        s1 = [1] * 10
        s2 = [2] * 4
        s3 = [3] * 10
        r = list(parallel_map(test_func, s1, s2, s3,
                              use_multiprocessing=False,
                              ordered=True))

        exp = [6] * 4
        self.assertEqual(r, exp)

    def test_multisequence_fill_void(self):
        def test_func(a, b, c):
            return a + b + c

        s1 = [1] * 10
        s2 = [2] * 4
        s3 = [3] * 10
        r = list(parallel_map(test_func, s1, s2, s3,
                              use_multiprocessing=False,
                              fill_void=10,
                              ordered=True))

        expected = [6] * 4 + [14] * 6
        self.assertEqual(r, expected)

    def test_nested_multiprocessing(self):
        # char -> char -> ord -> char -> ord
        g1 = parallel_map(lambda e: e, self.test_string,
                          ordered=True,
                          use_multiprocessing=True,
                          cores=2)
        g2 = parallel_map(ord, g1,
                          ordered=True,
                          use_multiprocessing=True,
                          cores=2)
        g3 = parallel_map(chr, g2,
                          ordered=True,
                          use_multiprocessing=True,
                          cores=2)
        g4 = parallel_map(ord, g3,
                          ordered=True,
                          use_multiprocessing=True,
                          cores=2)

        expected = list(map(ord, self.test_string))
        self.assertEqual(
            list(g4),
            expected
        )

    def test_nested_threading(self):
        # char -> char -> ord -> char -> ord
        g1 = parallel_map(lambda e: e, self.test_string,
                          ordered=True,
                          use_multiprocessing=False,
                          cores=2,
                          name='g1')
        g2 = parallel_map(ord, g1,
                          ordered=True,
                          use_multiprocessing=False,
                          cores=2,
                          name='g2')
        g3 = parallel_map(chr, g2,
                          ordered=True,
                          use_multiprocessing=False,
                          cores=2,
                          name='g3')
        g4 = parallel_map(ord, g3,
                          ordered=True,
                          use_multiprocessing=False,
                          cores=2,
                          name='g4')

        expected = list(map(ord, self.test_string))
        self.assertEqual(
            list(g4),
            expected
        )



from __future__ import division, print_function
import unittest

from smqtk.utils import merge_dict


class TestMergeDict (unittest.TestCase):

    def setUp(self):
        # Reset values
        self.a = {
            1: 2,
            3: 'value',
            'nested': {
                'v': 'sub-value',
                'l': [0, 1, 2],
                'i': 4,
                'even_deeper': {
                    'j': 'x',
                    'k': 'l',
                },
            },
            'nested2': {
                'a': 'will be overwritten'
            },
            's': {'some', 'set', 'here'}
        }

        self.b = {
            3: 'different_value',
            4: 'new value',
            'nested': {
                'l': [3, 4, 5],
                'i': 5,
                'new': 'pair',
                'even_deeper': {
                    'j': {
                        'new': 'sub-dict'
                    },
                },
            },
            'nested2': 'overwritten value',
        }

        self.expected = {
            1: 2,
            3: 'different_value',
            4: 'new value',
            'nested': {
                'v': 'sub-value',
                'l': [3, 4, 5],
                'i': 5,
                'new': 'pair',
                'even_deeper': {
                    'j': {
                        'new': 'sub-dict'
                    },
                    'k': 'l',
                },
            },
            'nested2': 'overwritten value',
            's': {'some', 'set', 'here'}
        }

    def test_merge_dict_shallow(self):
        # basic dict merger
        merge_dict(self.a, self.b)
        self.assertEqual(self.a, self.expected)

        # set values that are mutable structures should be the same instances as
        # what's in ``b``.
        self.assertEqual(self.a['nested']['l'],
                         self.b['nested']['l'])
        self.assertIs(self.a['nested']['l'],
                      self.b['nested']['l'])

        self.assertEqual(self.a['nested']['even_deeper']['j'],
                         self.b['nested']['even_deeper']['j'])
        self.assertIs(self.a['nested']['even_deeper']['j'],
                      self.b['nested']['even_deeper']['j'])

    def test_merge_dict_deepcopy(self):
        # dict merger with deepcopy
        merge_dict(self.a, self.b, deep_copy=True)
        self.assertEqual(self.a, self.expected)

        # set values that are mutable structures should be the same instances as
        # what's in ``b``.
        self.assertEqual(self.a['nested']['l'],
                         self.b['nested']['l'])
        self.assertIsNot(self.a['nested']['l'],
                         self.b['nested']['l'])

        self.assertEqual(self.a['nested']['even_deeper']['j'],
                         self.b['nested']['even_deeper']['j'])
        self.assertIsNot(self.a['nested']['even_deeper']['j'],
                         self.b['nested']['even_deeper']['j'])

    def test_merge_dict_return_a(self):
        # Return value should be the ``a`` parameter input value
        r = merge_dict(self.a, self.b)
        self.assertIs(r, self.a)



import logging
import hashlib
import multiprocessing
import os
import re
import shutil
import subprocess
import time
import six

import six

from smqtk.utils import file_utils, string_utils


__author__ = "paul.tunison@kitware.com"


class VideoMetadata (object):
    """
    Simple container for video file metadata values
    """

    def __init__(self):
        #: :type: None or int
        self.width = None
        #: :type: None or int
        self.height = None
        #: :type: None or float
        self.fps = None
        #: :type: None or float
        self.duration = None


def get_metadata_info(video_filepath, ffprobe_exe='ffprobe'):
    """
    Use ffmpeg to extract video file metadata parameters

    :param video_filepath: File path to the video to probe.
    :type video_filepath: str

    :param ffprobe_exe: Path to the ffprobe executable to use. By default, we
        try to use the version that's on the PATH.

    :return: VideoMetadata instance
    :rtype: VideoMetadata

    """
    log = logging.getLogger('smqtk.utils.video_utils.get_metadata_info')
    re_float_match = "[+-]?(?:(?:\d+\.?\d*)|(?:\.\d+))(?:[eE][+-]?\d+)?"

    log.debug("Using ffprobe: %s", ffprobe_exe)
    cmd = [ffprobe_exe, '-i', video_filepath]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    # ffprobe puts output to err stream
    if p.returncode:  # non-zero
        raise RuntimeError("Failed to probe video file. Error:\n%s"
                           % err)

    # WxH
    m = re.search("Stream.*Video.* (\d+)x(\d+)", err)
    if m:
        width = int(m.group(1))
        height = int(m.group(2))
    else:
        raise RuntimeError("Couldn't find width/height specification "
                           "for video file '%s'" % video_filepath)

    # FPS
    m = re.search("Stream.*Video.* (%s) fps" % re_float_match, err)
    if m:
        fps = float(m.group(1))
    else:
        # falling back on tbr measurement
        log.debug("Couldn't find fps measurement, looking for TBR")
        m = re.search("Stream.*Video.* (%s) tbr" % re_float_match, err)
        if m:
            fps = float(m.group(1))
        else:
            raise RuntimeError("Couldn't find tbr specification for "
                               "video file '%s'" % video_filepath)

    # Duration
    m = re.search("Duration: (\d+):(\d+):(%s)" % re_float_match, err)
    if m:
        duration = (
            (60 * 60 * int(m.group(1)))     # hours
            + (60 * int(m.group(2)))        # minutes
            + float(m.group(3))             # seconds
        )
    else:
        raise RuntimeError("Couldn't find duration specification for "
                           "video file '%s'" % video_filepath)

    md = VideoMetadata()
    md.width = width
    md.height = height
    md.fps = fps
    md.duration = duration

    return md


def ffmpeg_extract_frame(t, input_filepath, output_filepath,
                         ffmpeg_exe='ffmpeg'):
    """
    Extract a frame a the given time ``t`` from the input video file.
    Output file may not exist or be of 0 size if we failed to extract the frame.
    """
    cmd = [ffmpeg_exe, '-accurate_seek', '-ss', str(t), '-i', input_filepath,
           '-frames:v', '1', output_filepath]
    sPIPE = subprocess.PIPE
    p = subprocess.Popen(cmd, stdout=sPIPE, stderr=sPIPE)
    _, _ = p.communicate()
    # if p.returncode != 0:
    #     raise RuntimeError("FFmpeg failed to extract frame at time %f! "
    #                        "(return code: %d)"
    #                        % (t, p.returncode))


def ffmpeg_extract_frame_map(working_dir, video_filepath, second_offset=0,
                             second_interval=0, max_duration=0, frames=(),
                             output_image_ext="png", parallel=None,
                             ffmpeg_exe='ffmpeg'):
    """
    Return a mapping of video frame index to image file in the given output
    format.

    If frames requested have not yet been extracted (based on what's contained
    in the specified output directory), they are done now. This means that this
    method could take a little time to complete if there are many frames in the
    video file and this is the first time this is being called.

    This may return an empty list if there are no frames in the video for
    the specified, or default, constraints.

    Extracted frames are cached in a directory structure under the provided
    ``working_dir`` directory path: ``<working_dir>/VideoFrameExtraction``.
    Frames are extracted into separate directories based on the SHA1 checksum of
    the video file.

    :raises RuntimeError: No frames were extracted.

    :param working_dir: Working directory for frame extraction to occur in.
    :type working_dir: str

    :param video_filepath: Path to the video to extract frames from.
    :type video_filepath: str

    :param second_offset: Seconds into the video to start extracting
    :type second_offset: float

    :param second_interval: Number of seconds between extracted frames
    :type second_interval: float

    :param max_duration: Maximum number of seconds worth of extracted frames
        (starting from the specified offset). If <=0, we extract until the end
        of the video.
    :type max_duration: float

    :param frames: Specific exact frame numbers within the video to extract.
        Providing explicit frames causes offset, interval and duration
        parameters to be ignored and only the frames specified here to be
        extracted and returned.
    :type frames: collections.Iterable[int]

    :param output_image_ext: Extension to use for output images.
    :type output_image_ext: str

    :param parallel: Number of processes to use for frame extraction. This is
        None by default, meaning that all available cores/threads are used.
    :type parallel: int or None

    :param ffmpeg_exe: ffmpeg executable to use for frame extraction. By
        default, we attempt to use what is available of the PATH.
    :type ffmpeg_exe: str or unicode

    :return: Map of frame-to-filepath for requested video frames
    :rtype: dict[int, str]

    """
    log = logging.getLogger('smqtk.utils.video_utils.extract_frame_map')

    video_md = get_metadata_info(video_filepath)
    video_sha1sum = hashlib.sha1(open(video_filepath, 'rb').read()).hexdigest()
    frame_output_dir = os.path.join(
        working_dir,
        "VideoFrameExtraction",
        *string_utils.partition_string(video_sha1sum, 10)
        # 40 hex chars split into chunks of 4
    )
    file_utils.safe_create_dir(frame_output_dir)

    def filename_for_frame(frame, ext):
        """
        method standard filename for a given frame file
        """
        return "%08d.%s" % (frame, ext.lstrip('.'))

    def iter_frames_for_interval():
        """
        Return a generator expression yielding frame numbers from the input
        video that match the given query parameters. Indices returned are
        0-based (i.e. first frame is 0, not 1).

        We are making a sensible assumption that we are not dealing with frame
        speeds of over 1000Hz and rounding frame frame times to the neared
        thousandth of a second to mitigate floating point error.

        :rtype: list of int

        """
        _log = logging.getLogger('smqtk.utils.video_utils.extract_frame_map'
                                 '.iter_frames_for_interval')
        num_frames = int(video_md.fps * video_md.duration)
        first_frame = second_offset * video_md.fps
        _log.debug("First frame: %f", first_frame)
        if max_duration > 0:
            cutoff_frame = min(float(num_frames),
                               (max_duration + second_offset) * video_md.fps)
        else:
            cutoff_frame = float(num_frames)
        _log.debug("Cutoff frame: %f", cutoff_frame)
        if second_interval:
            incr = second_interval * video_md.fps
        else:
            incr = 1.0
        _log.debug("Frame increment: %f", incr)

        # Interpolate
        yield first_frame
        next_frm = first_frame + incr
        while next_frm < cutoff_frame:
            _log.debug("-- adding frame: %f", next_frm)
            yield int(next_frm)
            next_frm += incr

    # noinspection PyShadowingNames
    def extract_frames(frames_to_process):
        """
        Extract specific frames from the input video file using ffmpeg. If not
        all frames could be extracted, we return what we were able to extract.

        :param frames_to_process: Mapping of frame-number:filepath pairs to
            extract from the input video.
        :type frames_to_process: dict[int,str or unicode]

        :return: List of frames that were successfully extracted.
        :rtype: list[int]

        """
        _log = logging.getLogger('smqtk.utils.video_utils.extract_frame_map'
                                 '.extract_frames')

        # Setup temp extraction directory
        tmp_extraction_dir = os.path.join(frame_output_dir, ".TMP")
        if os.path.isdir(tmp_extraction_dir):
            _log.debug("Existing temp director found, removing and starting "
                       "over")
            shutil.rmtree(tmp_extraction_dir, ignore_errors=True)
        os.makedirs(tmp_extraction_dir)

        p = multiprocessing.Pool(parallel)
        # Mapping of frame to (result, output_filepath)
        #: :type: dict[int, (AsyncResult, str)]
        rmap = {}
        for f, ofp in six.iteritems(frames_to_process):
            tfp = os.path.join(tmp_extraction_dir,
                               filename_for_frame(f, output_image_ext))
            t = f / video_md.fps
            rmap[f] = (
                p.apply_async(ffmpeg_extract_frame,
                              args=(t, video_filepath, tfp, ffmpeg_exe)),
                tfp
            )
        p.close()
        # Check for failures
        extracted_frames = []
        for f, ofp in six.iteritems(frames_to_process):
            r, tfp = rmap[f]
            r.get()  # wait for finish
            if not os.path.isfile(tfp):
                _log.warn("Failed to generated file for frame %d", f)
            else:
                extracted_frames.append(f)
                os.rename(tfp, ofp)
        p.join()
        del p

        os.removedirs(tmp_extraction_dir)
        _log.debug("Frame extraction complete")

        return extracted_frames

    # Determine frames to extract from video
    extract_indices = set()
    if frames:
        log.debug("Only extracting specified frames: %s", frames)
        extract_indices.update(frames)
    else:
        log.debug("Determining frames needed for specification: "
                  "offset: %f, interval: %f, max_duration: %f",
                  second_offset, second_interval, max_duration)
        extract_indices.update(iter_frames_for_interval())

    if not extract_indices:
        return {}

    # frame/filename map that will be returned based on requested frames
    frame_map = dict(
        (i, os.path.join(frame_output_dir,
                         filename_for_frame(i, output_image_ext)))
        for i in extract_indices
    )

    ###
    # Acquire a file-base lock in output directory so that we don't conflict
    # with another process extracting frames to the same directory.
    #
    # NOTE: This method is prone to starvation if many processes are trying
    #       to extract to the same video frames, but not yet probably due to
    #       existing use cases.
    #
    lock_file = os.path.join(frame_output_dir, '.lock')
    log.debug("Acquiring file lock in '%s'...", frame_output_dir)
    while not file_utils.exclusive_touch(lock_file):
        # This is sufficiently small to be fine grained, but not so small to
        # burn the CPU.
        time.sleep(0.01)
    log.debug("Acquiring file lock -> Acquired!")

    try:
        ###
        # Determine frames to actually extract base on existing files (if any)
        #
        #: :type: dict[int, str]
        frames_to_process = {}
        existing_frames = []
        for i, img_file in sorted(frame_map.items()):
            if not os.path.isfile(img_file):
                log.debug('frame %d needs processing', i)
                frames_to_process[i] = img_file
            else:
                existing_frames.append(i)

        ###
        # Extract needed frames via hook function that provides
        # implementation.
        #
        if frames_to_process:
            frames_extracted = extract_frames(frames_to_process)

            if (len(existing_frames) + len(frames_extracted)) == 0:
                raise RuntimeError("Failed to extract any frames for video")

        return frame_map
    finally:
        os.remove(lock_file)



import collections
import copy
import os.path as osp

import numpy
import six
from six.moves import range, zip
from six.moves import cPickle as pickle

from smqtk.algorithms.relevancy_index import RelevancyIndex
from smqtk.utils.distance_kernel import (
    compute_distance_matrix
)
from smqtk.utils.metrics import histogram_intersection_distance
from smqtk.utils.parallel import parallel_map

try:
    import svm
    import svmutil
except ImportError:
    svm = None
    svmutil = None


class LibSvmHikRelevancyIndex (RelevancyIndex):
    """
    Uses libSVM python interface, using histogram intersection, to implement
    IQR ranking.
    """

    # Dictionary of parameter/value pairs that will be passed to libSVM during
    # the model trail phase. Parameters that are flags, i.e. have no values,
    # should be given an empty string ('') value.
    SVM_TRAIN_PARAMS = {
        '-q': '',
        '-t': 5,  # Specified the use of HI kernel, 5 is unique to custom build
        '-b': 1,
        '-c': 2,
        '-g': 0.0078125,
    }

    @classmethod
    def is_usable(cls):
        """
        Check whether this implementation is available for use.

        Required valid presence of svm and svmutil modules

        :return:
            Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        return svm and svmutil

    def __init__(self, descr_cache_filepath=None, autoneg_select_ratio=1,
                 multiprocess_fetch=False, cores=None):
        """
        Initialize a new or existing index.

        TODO ::
        - input optional known background descriptors, i.e. descriptors for
            things that would otherwise always be considered a negative example.

        :param descr_cache_filepath: Optional path to store/load descriptors
            we index.
        :type descr_cache_filepath: None | str

        :param autoneg_select_ratio: Number of maximally distant descriptors to
            select from our descriptor cache for each positive example provided
            when no negative examples are provided for ranking.

            This must be an integer. Default value of 1. It is advisable not to
            make this value too large.
        :type autoneg_select_ratio: int

        :param multiprocess_fetch: Use multiprocessing vs threading when
            fetching descriptor vectors during ``build_index``. Default is
            False.
        :type multiprocess_fetch: bool

        :param cores: Cores to use when performing parallel operations. A value
            of None means to use all available cores.
        :type cores: int | None

        """
        super(LibSvmHikRelevancyIndex, self).__init__()

        self.descr_cache_fp = descr_cache_filepath
        self.autoneg_select_ratio = int(autoneg_select_ratio)
        self.multiprocess_fetch = multiprocess_fetch
        self.cores = cores

        # Descriptor elements in this index
        self._descr_cache = []
        # Local serialization of descriptor vectors. Used when for computing
        # distances of SVM support vectors for Platt Scaling
        self._descr_matrix = None
        # Mapping of descriptor vectors to their index in the cache, and
        # subsequently in the distance kernel
        self._descr2index = {}
        # # Distance kernel matrix (symmetric)
        # self._dist_kernel = None

        if self.descr_cache_fp and osp.exists(self.descr_cache_fp):
            with open(self.descr_cache_fp, 'rb') as f:
                descriptors = pickle.load(f)
                # Temporarily unsetting so we don't cause an extra write inside
                # build_index.
                self.descr_cache_fp = None
                self.build_index(descriptors)
                self.descr_cache_fp = descr_cache_filepath

    @staticmethod
    def _gen_w1_weight(num_pos, num_neg):
        """
        Return w1 weight parameter based on pos and neg exemplars
        """
        return max(1.0, num_neg/float(num_pos))

    @classmethod
    def _gen_svm_parameter_string(cls, num_pos, num_neg):
        params = copy.copy(cls.SVM_TRAIN_PARAMS)
        params['-w1'] = cls._gen_w1_weight(num_pos, num_neg)
        return ' '.join(('%s %s' % (k, v) for k, v in params.items()))

    def get_config(self):
        return {
            "descr_cache_filepath": self.descr_cache_fp,
            'autoneg_select_ratio': self.autoneg_select_ratio,
            'multiprocess_fetch': self.multiprocess_fetch,
            'cores': self.cores,
        }

    def count(self):
        return len(self._descr_cache)

    def build_index(self, descriptors):
        """
        Build the index based on the given iterable of descriptor elements.

        Subsequent calls to this method should rebuild the index, not add to
        it.

        :raises ValueError: No data available in the given iterable.

        :param descriptors:
            Iterable of descriptor elements to build index over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        # ordered cache of descriptors in our index.
        self._descr_cache = []
        # Reverse mapping of a descriptor's vector to its index in the cache
        # and subsequently in the distance kernel.
        self._descr2index = {}
        # matrix for creating distance kernel
        self._descr_matrix = []

        def get_vector(d_elem):
            return d_elem, d_elem.vector()

        # noinspection PyTypeChecker
        vector_iter = parallel_map(get_vector, descriptors,
                                   name='vector_iter',
                                   use_multiprocessing=self.multiprocess_fetch,
                                   cores=self.cores,
                                   ordered=True)

        for i, (d, v) in enumerate(vector_iter):
            self._descr_cache.append(d)
            # ``_descr_matrix`` is a list, currently.
            # noinspection PyUnresolvedReferences
            self._descr_matrix.append(v)
            self._descr2index[tuple(v)] = i
        self._descr_matrix = numpy.array(self._descr_matrix)

        # TODO: (?) For when we optimize SVM SV kernel computation
        # self._dist_kernel = \
        #    compute_distance_kernel(self._descr_matrix,
        #                            histogram_intersection_distance2,
        #                            row_wise=True)

        if self.descr_cache_fp:
            with open(self.descr_cache_fp, 'wb') as f:
                pickle.dump(self._descr_cache, f, -1)

    def rank(self, pos, neg):
        """
        Rank the currently indexed elements given ``pos`` positive and ``neg``
        negative exemplar descriptor elements.

        :param pos: Iterable of positive exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type pos: collections.Iterable[smqtk.representation.DescriptorElement]

        :param neg: Iterable of negative exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type neg: collections.Iterable[smqtk.representation.DescriptorElement]

        :return: Map of indexed descriptor elements to a rank value between
            [0, 1] (inclusive) range, where a 1.0 means most relevant and 0.0
            meaning least relevant.
        :rtype: dict[smqtk.representation.DescriptorElement, float]

        """
        # Notes:
        # - Pos and neg exemplars may be in our index.

        #
        # SVM model training
        #
        # Copy pos descriptors into a set for repeated iteration
        #: :type: set[smqtk.representation.DescriptorElement]
        pos = set(pos)
        # Creating training matrix and labels
        train_labels = []
        train_vectors = []
        num_pos = 0
        for d in pos:
            train_labels.append(+1)
            train_vectors.append(d.vector().tolist())
            num_pos += 1
        self._log.debug("Positives given: %d", num_pos)

        # When no negative examples are given, naively pick most distant
        # example in our dataset, using HI metric, for each positive example
        neg_autoselect = set()
        # Copy neg descriptors into a set for testing size.
        if not isinstance(neg, collections.Sized):
            #: :type: set[smqtk.representation.DescriptorElement]
            neg = set(neg)
        if not neg:
            self._log.info("Auto-selecting negative examples. (%d per "
                           "positive)", self.autoneg_select_ratio)
            # ``train_vectors`` only composed of positive examples at this
            # point.
            for p in pos:
                # Where d is the distance vector to descriptor elements in
                # cache.
                d = histogram_intersection_distance(p.vector(),
                                                    self._descr_matrix)
                # Scan vector for max distance index
                # - Allow variable number of maximally distance descriptors to
                #   be picked per positive.
                # track most distance neighbors
                m_set = {}
                # track smallest distance of most distant neighbors
                m_val = -float('inf')
                for i in range(d.size):
                    if d[i] > m_val:
                        m_set[d[i]] = i
                        if len(m_set) > self.autoneg_select_ratio:
                            if m_val in m_set:
                                del m_set[m_val]
                        m_val = min(m_set)
                for i in six.itervalues(m_set):
                    neg_autoselect.add(self._descr_cache[i])
            # Remove any positive examples from auto-selected results
            neg_autoselect.difference_update(pos)
            self._log.debug("Auto-selected negative descriptors [%d]: %s",
                            len(neg_autoselect), neg_autoselect)

        num_neg = 0
        for n_iterable in (neg, neg_autoselect):
            for d in n_iterable:
                train_labels.append(-1)
                train_vectors.append(d.vector().tolist())
                num_neg += 1

        if not num_pos:
            raise ValueError("No positive examples provided.")
        elif not num_neg:
            raise ValueError("No negative examples provided.")

        # Training SVM model
        self._log.debug("online model training")
        svm_problem = svm.svm_problem(train_labels, train_vectors)
        param_str = self._gen_svm_parameter_string(num_pos, num_neg)
        svm_param = svm.svm_parameter(param_str)
        svm_model = svmutil.svm_train(svm_problem, svm_param)

        if hasattr(svm_model, "param"):
            self._log.debug("SVM input parameters: %s", param_str)
            self._log.debug("SVM model parsed parameters: %s",
                            str(svm_model.param))
            param = svm_model.param
            wgt_pairs = [(param.weight_label[i], param.weight[i])
                         for i in range(param.nr_weight)]
            wgt_str = " ".join(["%s: %s" % wgt for wgt in wgt_pairs])
            self._log.debug("SVM model parsed weight parameters: %s", wgt_str)

        if svm_model.l == 0:
            raise RuntimeError("SVM Model learning failed")

        #
        # Platt Scaling for probability rankings
        #

        self._log.debug("making test distance matrix")
        # Number of support vectors
        # Q: is this always the same as ``svm_model.l``?
        num_SVs = sum(svm_model.nSV[:svm_model.nr_class])
        # Support vector dimensionality
        dim_SVs = len(train_vectors[0])
        # initialize matrix they're going into
        svm_SVs = numpy.ndarray((num_SVs, dim_SVs), dtype=float)
        for i, nlist in enumerate(svm_model.SV[:svm_SVs.shape[0]]):
            svm_SVs[i, :] = [n.value for n in nlist[:len(train_vectors[0])]]
        # compute matrix of distances from support vectors to index elements
        # TODO: Optimize this step by caching SV distance vectors
        #       - It is known that SVs are vectors from the training data, so
        #           if the same descriptors are given to this function
        #           repeatedly (which is the case for IQR), this can be faster
        #           because we're only computing at most a few more distance
        #           vectors against our indexed descriptor matrix, and the rest
        #           have already been computed before.
        #       - At worst, we're effectively doing this call because each SV
        #           needs to have its distance vector computed.
        svm_test_k = compute_distance_matrix(svm_SVs, self._descr_matrix,
                                             histogram_intersection_distance,
                                             row_wise=True)

        # TODO(john.moeller): None of the Platt scaling should be necessary.
        # svmutil.svm_predict will apply the Platt scaling directly. See
        # https://github.com/cjlin1/libsvm/tree/master/python

        self._log.debug("Platt scaling")
        # the actual platt scaling stuff
        weights = numpy.array(svm_model.get_sv_coef()).flatten()
        margins = numpy.dot(weights, svm_test_k)
        rho = svm_model.rho[0]
        probA = svm_model.probA[0]
        probB = svm_model.probB[0]
        #: :type: numpy.core.multiarray.ndarray
        probs = 1.0 / (1.0 + numpy.exp((margins - rho) * probA + probB))

        # Detect whether we need to flip probabilities
        # - Probability of input positive examples should have a high
        #   probability score among the generated probabilities of our index.
        # - If the positive example probabilities show to be in the lower 50%,
        #   flip the generated probabilities, since its experimentally known
        #   that the SVM will change which index it uses to represent a
        #   particular class label occasionally, which influences the Platt
        #   scaling apparently.
        pos_vectors = numpy.array(train_vectors[:num_pos])
        pos_test_k = compute_distance_matrix(svm_SVs, pos_vectors,
                                             histogram_intersection_distance,
                                             row_wise=True)
        pos_margins = numpy.dot(weights, pos_test_k)
        #: :type: numpy.core.multiarray.ndarray
        pos_probs = 1.0 / (1.0 + numpy.exp((pos_margins - rho) * probA + probB))
        # Check if average positive probability is less than the average index
        # probability. If so, the platt scaling probably needs to be flipped.
        if (pos_probs.sum() / pos_probs.size) < (probs.sum() / probs.size):
            self._log.debug("inverting probabilities")
            probs = 1. - probs

        rank_pool = dict(zip(self._descr_cache, probs))
        return rank_pool


RELEVANCY_INDEX_CLASS = LibSvmHikRelevancyIndex



import mock
import os
import requests
import unittest

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation.data_element.url_element import DataUrlElement
from smqtk.tests import TEST_DATA_DIR


# Check to see if we have an internet connection.
internet_available = True
try:
    r = requests.get('https://data.kitware.com')
    _ = r.content
except requests.ConnectionError:
    internet_available = False


class TestDataUrlElement (unittest.TestCase):
    """

    :NOTE: Some of these tests require a connection to the internet in order to
        pass.

    """

    # Public domain Lenna image from Wikipedia, same as local test image file
    EXAMPLE_URL = \
        'https://data.kitware.com/api/v1/file/5820bbeb8d777f10f26efc2f/download'
    EXAMPLE_PTH = os.path.join(TEST_DATA_DIR, 'Lenna.png')

    def test_is_usable(self):
        # Should always be available, because local/intranet networks are a
        # thing.
        self.assertTrue(DataUrlElement.is_usable())

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_new_from_internet(self):
        e = DataUrlElement(self.EXAMPLE_URL)
        self.assertEqual(e.get_bytes(), open(self.EXAMPLE_PTH, 'rb').read())

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_new_add_missing_scheme(self):
        # Construct without scheme header, should add http://
        e = DataUrlElement(self.EXAMPLE_URL[8:])
        self.assertEqual(e._url, 'http://' + self.EXAMPLE_URL[8:])
        self.assertEqual(e.get_bytes(), open(self.EXAMPLE_PTH, 'rb').read())

    def test_new_invalid_url(self):
        self.assertRaises(
            requests.ConnectionError,
            DataUrlElement,
            'http://not.a.real.host/'
        )

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_content_type(self):
        e = DataUrlElement(self.EXAMPLE_URL)
        self.assertEqual(e.content_type(), 'image/png')

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_from_configuration(self):
        default_config = DataUrlElement.get_default_config()
        self.assertEqual(default_config, {'url_address': None})

        default_config['url_address'] = self.EXAMPLE_URL
        #: :type: DataUrlElement
        inst1 = DataUrlElement.from_config(default_config)
        self.assertEqual(default_config, inst1.get_config())
        self.assertEqual(inst1._url, self.EXAMPLE_URL)
        self.assertEqual(inst1.get_bytes(), open(self.EXAMPLE_PTH, 'rb').read())

        inst2 = DataUrlElement.from_config(inst1.get_config())
        self.assertEqual(inst1, inst2)

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_from_uri(self):
        e = DataUrlElement.from_uri(self.EXAMPLE_URL)
        self.assertEqual(e.get_bytes(), open(self.EXAMPLE_PTH, 'rb').read())
        self.assertEqual(e.content_type(), 'image/png')

    def test_from_uri_no_scheme(self):
        self.assertRaises(
            InvalidUriError,
            DataUrlElement.from_uri,
            'www.kitware.com'
        )

    def test_from_uri_invalid_scheme(self):
        self.assertRaises(
            InvalidUriError,
            DataUrlElement.from_uri,
            'ftp://www.kitware.com'
        )

    @mock.patch('smqtk.representation.data_element.url_element.requests.get')
    def test_is_empty_zero_bytes(self, _m_requests_get):
        e = DataUrlElement('some-address')
        # simulate no content bytes returned
        e.get_bytes = mock.MagicMock(return_value='')
        self.assertTrue(e.is_empty())

    @mock.patch('smqtk.representation.data_element.url_element.requests.get')
    def test_is_empty_nonzero_bytes(self, _m_requests_get):
        e = DataUrlElement('some-address')
        # simulate some content bytes returned
        e.get_bytes = mock.MagicMock(return_value='some bytes returned')
        self.assertFalse(e.is_empty())

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_get_bytes_from_url(self):
        e = DataUrlElement(self.EXAMPLE_URL)
        self.assertEqual(e.get_bytes(), open(self.EXAMPLE_PTH, 'rb').read())
        self.assertEqual(e.content_type(), 'image/png')

    @mock.patch('smqtk.representation.data_element.url_element.requests.get')
    def test_get_bytes_404_return_code(self, m_requests_get):
        e = DataUrlElement('some-address')

        sim_rc = 500
        simulated_r = requests.Response()
        simulated_r.status_code = sim_rc
        m_requests_get.return_value = simulated_r

        self.assertRaisesRegexp(
            requests.HTTPError,
            '%d' % sim_rc,
            e.get_bytes
        )

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_writable_universal_false(self):
        # Cannot write to a URL hosted file, should always return false
        e = DataUrlElement(self.EXAMPLE_URL)
        self.assertFalse(e.writable())

    @unittest.skipUnless(internet_available, "Internet not accessible")
    def test_set_bytes_universal_readonly(self):
        # should never be able to write to a url element
        e = DataUrlElement(self.EXAMPLE_URL)
        self.assertRaises(
            ReadOnlyError,
            e.set_bytes, 'foo'
        )



"""
Example where only explicitly listed classes are exported even though others are
implemented.
"""

from smqtk.tests.utils.test_plugin_get import DummyInterface


class ImplDoExport (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'doExport'+str(val)


class ImplNoExport (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'noExport'+str(val)


TEST_PLUGIN_CLASS = ImplDoExport



import mimetypes
import re
import requests

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation import DataElement


MIMETYPES = mimetypes.MimeTypes()


class DataUrlElement (DataElement):
    """
    Representation of data loadable via a web URL address.
    """

    # Enforce presence of demarcating schema
    URI_RE = re.compile('^https?://.+$')

    @classmethod
    def is_usable(cls):
        # URLs are not necessarily on the public internet. Local networking
        # should always be available.
        return True

    @classmethod
    def from_uri(cls, uri):
        m = cls.URI_RE.match(uri)
        if m is not None:
            # simply pass on URI as URL address
            return DataUrlElement(uri)

        raise InvalidUriError(uri, "Invalid web URI")

    def __init__(self, url_address):
        """
        Create a new URL element for a URL address.

        The given address may not resolve to anything.

        :raises requests.exceptions.ConnectionError: Failed to connect with the
            given hostname.
        :raises requests.exceptions.HTTPError: URL address provided does not
            resolve into a valid GET request.

        :param url_address: Web address of element
        :type url_address: str

        """
        super(DataUrlElement, self).__init__()

        self._url = url_address

        # make sure that url has a http:// or https:// prefix
        if not (self._url[:7] == "http://" or self._url[:8] == "https://"):
            self._url = "http://" + self._url

        # Check that the URL is valid, i.e. actually points to something
        requests.get(self._url).raise_for_status()

    def __repr__(self):
        return super(DataUrlElement, self).__repr__() + "{url: %s}" % self._url

    def get_config(self):
        return {
            "url_address": self._url
        }

    def content_type(self):
        """
        :return: Standard type/subtype string for this data element, or None if
            the content type is unknown.
        :rtype: str or None
        """
        return requests.get(self._url).headers['content-type']

    def is_empty(self):
        """
        Check if this element contains no bytes.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """
        return len(self.get_bytes()) == 0

    def get_bytes(self):
        """
        :return: Get the byte stream for this data element.
        :rtype: bytes

        :raises requests.exceptions.HTTPError: Error during request for data
            via GET.

        """
        # Fetch content from URL, return bytes
        r = requests.get(self._url)
        r.raise_for_status()
        return r.content

    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        # Web addresses cannot be written to
        return False

    def set_bytes(self, b):
        """
        Set bytes to this data element in the form of a string.

        Not all implementations may support setting bytes (writing). See the
        ``writable`` method.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        raise ReadOnlyError("URL address targets cannot be written to.")


DATA_ELEMENT_CLASS = DataUrlElement



import abc
import os

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.utils.plugin import get_plugins


__author__ = "paul.tunison@kitware.com"


class RelevancyIndex (SmqtkAlgorithm):
    """
    Abstract class for IQR index implementations.

    Similar to a traditional nearest-neighbors algorithm, An IQR index provides
    a specialized nearest-neighbors interface that can take multiple examples of
    positively and negatively relevant exemplars in order to produce a [0, 1]
    ranking of the indexed elements by determined relevancy.

    """

    def __len__(self):
        return self.count()

    @abc.abstractmethod
    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """

    @abc.abstractmethod
    def build_index(self, descriptors):
        """
        Build the index based on the given iterable of descriptor elements.

        Subsequent calls to this method should rebuild the index, not add to it.

        :raises ValueError: No data available in the given iterable.

        :param descriptors: Iterable of descriptor elements to build index over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """

    @abc.abstractmethod
    def rank(self, pos, neg):
        """
        Rank the currently indexed elements given ``pos`` positive and ``neg``
        negative exemplar descriptor elements.

        :param pos: Iterable of positive exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type pos: collections.Iterable[smqtk.representation.DescriptorElement]

        :param neg: Iterable of negative exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type neg: collections.Iterable[smqtk.representation.DescriptorElement]

        :return: Map of indexed descriptor elements to a rank value between
            [0, 1] (inclusive) range, where a 1.0 means most relevant and 0.0
            meaning least relevant.
        :rtype: dict[smqtk.representation.DescriptorElement, float]

        """


def get_relevancy_index_impls(reload_modules=False):
    """
    Discover and return discovered ``RelevancyIndex`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable
          ``RELEVANCY_INDEX_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``RELEVANCY_INDEX_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``RelevancyIndex``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "RELEVANCY_INDEX_PATH"
    helper_var = "RELEVANCY_INDEX_CLASS"
    return get_plugins(__name__, this_dir, env_var, helper_var, RelevancyIndex,
                       reload_modules=reload_modules)



import os

from smqtk.utils.plugin import get_plugins
from ._interface_nn_index import NearestNeighborsIndex


__all__ = [
    'NearestNeighborsIndex', 'get_nn_index_impls',
]


def get_nn_index_impls(reload_modules=False):
    """
    Discover and return discovered ``NearestNeighborsIndex`` classes. Keys in
    the returned map are the names of the discovered classes, and the paired
    values are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that
          begin with an alphanumeric character),
        - python modules listed in the environment variable ``NN_INDEX_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH
              separator character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``NN_INDEX_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``NearestNeighborsIndex``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "NN_INDEX_PATH"
    helper_var = "NN_INDEX_CLASS"
    return get_plugins(__name__, this_dir, env_var, helper_var,
                       NearestNeighborsIndex, reload_modules=reload_modules)



"""
Helper interface and functions for higher level plugin module getter methods.

Plugin configuration dictionaries take the following general format:

.. code-block:: json

    {
        "type": "name",
        "impl_name": {
            "param1": "val1",
            "param2": "val2"
        },
        "other_impl": {
            "p1": 4.5,
            "p2": null
        }
    }

"""

import abc
import collections
import importlib
import logging
import os
import re
import sys

import six
from six.moves import reload_module


# Template for checking validity of sub-module files
VALID_MODULE_FILE_RE = re.compile("^[a-zA-Z]\w*(?:\.py)?$")

# Template for checking validity of module attributes
VALUE_ATTRIBUTE_RE = re.compile("^[a-zA-Z]\w*$")

OS_ENV_PATH_SEP = (sys.platform == "win32" and ';') or ':'


class NotUsableError (Exception):
    """
    Exception thrown when a pluggable class is constructed but does not report
    as usable.
    """


@six.add_metaclass(abc.ABCMeta)
class Pluggable (object):
    """
    Interface for classes that have plugin implementations
    """

    @classmethod
    @abc.abstractmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        Since certain plugin implementations may require additional dependencies
        that may not yet be available on the system, this method should check
        for those dependencies and return a boolean saying if the implementation
        is usable.

        NOTES:
            - This should be a class method
            - When an implementation is deemed not usable, this should emit a
                warning detailing why the implementation is not available for
                use.

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        raise NotImplementedError("is_usable class-method not implemented for "
                                  "class '%s'" % cls.__name__)

    def __init__(self):
        if not self.is_usable():
            raise NotUsableError("Implementation class '%s' is not currently "
                                 "usable." % self.__class__.__name__)


def get_plugins(base_module_str, internal_dir, dir_env_var, helper_var,
                baseclass_type, warn=True, reload_modules=False):
    """
    Discover and return classes found in the SMQTK internal plugin directory and
    any additional directories specified via an environment variable.

    In order to specify additional out-of-SMQTK python modules containing
    base-class implementations, additions to the given environment variable must
    be made. Entries must be separated by either a ';' (for windows) or ':' (for
    everything else). This is the same as for the PATH environment variable on
    your platform. Entries should be paths to importable modules containing
    attributes for potential import.

    When looking at module attributes, we acknowledge those that start with an
    alphanumeric character ('_' prefixed attributes are hidden from import by
    this function).

    We required that the base class that we are checking for also descends from
    the ``Pluggable`` interface defined above. This allows us to check if a
    loaded class ``is_usable``.

    Within a module we first look for a helper variable by the name provided,
    which can either be a single class object or an iterable of class objects,
    to be specifically exported. If the variable is set to None, we skip that
    module and do not import anything. If the variable is not present, we look
    at attributes defined in that module for classes that descend from the given
    base class type. If none of the above are found, or if an exception occurs,
    the module is skipped.

    :param base_module_str: SMQTK internal string module path in which internal
        plugin modules are located.
    :type base_module_str: str

    :param internal_dir: Directory path to where SMQTK internal plugin modules
        are located.
    :type internal_dir: str

    :param dir_env_var: String name of an environment variable to look for that
        may optionally define additional directory paths to search for modules
        that may implement additional child classes of the base type.
    :type dir_env_var: str

    :param helper_var: Name of the expected module helper attribute.
    :type helper_var: str

    :param baseclass_type: Class type that discovered classes should descend
        from (inherit from).
    :type baseclass_type: type

    :param warn: If we should warn about module import failures.
    :type warn: bool

    :param reload_modules: Explicitly reload discovered modules from source
        instead of taking a potentially cached version of the module.
    :type reload_modules: bool

    :return: Map of discovered class objects descending from type
        ``baseclass_type`` and ``smqtk.utils.plugin.Pluggable`` whose keys are
        the string names of the class types.
    :rtype: dict[str, type]

    """
    log = logging.getLogger('.'.join([__name__,
                                      "getPlugins[%s]" % base_module_str]))

    if not issubclass(baseclass_type, Pluggable):
        raise ValueError("Required base-class must descend from the Pluggable "
                         "interface!")

    # List of module paths to check for valid sub-classes.
    #: :type: list[str]
    module_paths = []

    # modules nested under internal module
    log.debug("Finding internal modules...")
    for filename in os.listdir(internal_dir):
        if VALID_MODULE_FILE_RE.match(filename):
            module_name = os.path.splitext(filename)[0]
            log.debug("-- %s", module_name)
            module_paths.append('.'.join([base_module_str, module_name]))
    log.debug("Internal modules to search: %s", module_paths)

    # modules from env variable
    log.debug("Extracting env var module paths")
    log.debug("-- path sep: %s", OS_ENV_PATH_SEP)
    if dir_env_var in os.environ:
        env_var_module_paths = os.environ[dir_env_var].split(OS_ENV_PATH_SEP)
        # strip out empty strings
        env_var_module_paths = [p for p in env_var_module_paths if p]
        log.debug("Additional module paths specified in env var: %s",
                  env_var_module_paths)
        module_paths.extend(env_var_module_paths)
    else:
        log.debug("No paths added from environment.")

    log.debug("Getting plugins for module '%s'", base_module_str)
    class_map = {}
    for module_path in module_paths:
        log.debug("Examining module: %s", module_path)
        # We want any exception this might throw to continue up. If a module
        # in the directory is not importable, the user should know.
        try:
            _module = importlib.import_module(module_path)
        except Exception as ex:
            if warn:
                log.warn("[%s] Failed to import module due to exception: "
                         "(%s) %s",
                         module_path, ex.__class__.__name__, str(ex))
            continue
        if reload_modules:
            # Invoke reload in case the module changed between imports.
            # six should find the right thing.
            # noinspection PyCompatibility
            _module = reload_module(_module)
            if _module is None:
                raise RuntimeError("[%s] Failed to reload"
                                   % module_path)

        # Find valid classes in the discovered module by:
        classes = []
        if hasattr(_module, helper_var):
            # Looking for magic variable for import guidance
            classes = getattr(_module, helper_var)
            if classes is None:
                log.debug("[%s] Helper is None-valued, skipping module",
                          module_path)
                classes = []
            elif (isinstance(classes, collections.Iterable) and
                  not isinstance(classes, six.string_types)):
                classes = list(classes)
                log.debug("[%s] Loaded list of %d class types via helper",
                          module_path, len(classes))
            elif issubclass(classes, baseclass_type):
                log.debug("[%s] Loaded class type: %s",
                          module_path, classes.__name__)
                classes = [classes]
            else:
                raise RuntimeError("[%s] Helper variable set to an invalid "
                                   "value: %s" % (module_path, classes))
        else:
            # Scan module valid attributes for classes that descend from the
            # given base-class.
            log.debug("[%s] No helper, scanning module attributes",
                      module_path)
            for attr_name in dir(_module):
                if VALUE_ATTRIBUTE_RE.match(attr_name):
                    classes.append(getattr(_module, attr_name))

        # Check the validity of the discovered class types
        for cls in classes:
            # check that all class types in iterable are:
            # - Class types,
            # - Subclasses of the given base-type and plugin interface
            # - Not missing any abstract implementations.
            #
            # noinspection PyUnresolvedReferences
            if not isinstance(cls, type):
                # No logging, over verbose, undetermined type.
                pass
            elif cls is baseclass_type:
                log.debug("[%s.%s] [skip] Literally the base class.",
                          module_path, cls.__name__)
            elif not issubclass(cls, baseclass_type):
                log.debug("[%s.%s] [skip] Does not descend from base class.",
                          module_path, cls.__name__)
            elif bool(cls.__abstractmethods__):
                # Making this a warning as I think this indicates a broken
                # implementation in the ecosystem.
                # noinspection PyUnresolvedReferences
                log.warn('[%s.%s] [skip] Does not implement one or more '
                         'abstract methods: %s',
                         module_path, cls.__name__,
                         list(cls.__abstractmethods__))
            elif not cls.is_usable():
                log.debug("[%s.%s] [skip] Class does not report as usable.",
                          module_path, cls.__name__)
            else:
                log.debug('[%s.%s] [KEEP] Retaining subclass.',
                          module_path, cls.__name__)
                class_map[cls.__name__] = cls

    return class_map


def make_config(plugin_map):
    """
    Generated configuration dictionary for the given plugin getter method
    (which returns a dictionary of labels to class types)

    A types parameters, as listed, at the construction parameters for that
    type. Default values are inserted where possible, otherwise None values are
    used.

    :param plugin_map: A dictionary mapping class names to class types.
    :type plugin_map: dict[str, type]

    :return: Base configuration dictionary with an empty ``type`` field, and
        containing the types and initialization parameter specification for all
        implementation types available from the provided getter method.
    :rtype: dict[str, object]

    """
    d = {"type": None}
    for label, cls in six.iteritems(plugin_map):
        # noinspection PyUnresolvedReferences
        d[label] = cls.get_default_config()
    return d


def to_plugin_config(cp_inst):
    """
    Helper method that transforms the configuration dictionary gotten from the
    passed Configurable-subclass instance into the standard multi-plugin
    configuration dictionary format (see above).

    This result of this function would be compatible with being passed to the
    ``from_plugin_config`` function, given the appropriate plugin-getter method.

    TL;DR: This wraps the instance's ``get_config`` return in a certain way
           that's compatible with ``from_plugin_config``.

    :param cp_inst: Instance of a Configurable-subclass.
    :type cp_inst: Configurable

    :return: Plugin-format configuration dictionary.
    :rtype: dict

    """
    name = cp_inst.__class__.__name__
    return {
        "type": name,
        name: cp_inst.get_config()
    }


def from_plugin_config(config, plugin_map, *args):
    """
    Helper method for instantiating an instance of a class available via the
    provided ``plugin_getter`` function given the plugin configuration
    dictionary ``config``.

    :raises ValueError:
        This may be raised if:
            - type field set to ``None``
            - type field did not match any available configuration in the given
              config.
            - Type field did not specify any implementation key.

    :raises TypeError: Insufficient/incorrect initialization parameters were
        specified for the specified ``type``'s constructor.

    :param config: Configuration dictionary to draw from.
    :type config: dict

    :param plugin_map: A dictionary mapping class names to class types.
    :type plugin_map: dict[str, type]

    :param args: Additional argument to be passed to the ``from_config`` method
        on the configured class type.

    :return: Instance of the configured class type as found in the given
        ``plugin_getter``.
    :rtype: smqtk.utils.Configurable

    """
    if 'type' not in config:
        raise ValueError("Configuration dictionary given does not have an "
                         "implementation type specification.")
    t = config['type']
    config_type_options = set(config.keys()) - {'type'}
    plugin_type_options = set(plugin_map.keys())
    # Type provided may either by None, not have a matching block in the config,
    # not have a matching implementation type, or match both.
    if t is None:
        raise ValueError("No implementation type specified. Options: %s"
                         % config_type_options)
    elif t not in config_type_options:
        raise ValueError("Implementation type specified as '%s', but no "
                         "configuration block was present for that type. "
                         "Available configuration block options: %s"
                         % (t, list(config_type_options)))
    elif t not in plugin_type_options:
        raise ValueError("Implementation type specified as '%s', but no "
                         "plugin implementations are available for that type. "
                         "Available implementation types options: %s"
                         % (t, list(plugin_type_options)))
    #: :type: smqtk.utils.Configurable
    cls = plugin_map[t]
    # noinspection PyUnresolvedReferences
    return cls.from_config(config[t], *args)



import mimetypes
import os.path as osp
import re

import six

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation import DataElement
from smqtk.utils.file_utils import safe_file_write


STR_NONE_TYPES = six.string_types + (type(None),)


class DataFileElement (DataElement):
    """
    File-based data element
    """

    # Regex for file paths optionally including the file:// and / prefix.
    # Allows any character between slashes currently.
    FILE_URI_RE = re.compile("^(?:file://)?(/?[^/]+(?:/[^/]+)*)$")

    @classmethod
    def is_usable(cls):
        # No dependencies
        return True

    @classmethod
    def from_uri(cls, uri):
        """
        Construct a new instance based on the given URI.

        File elements can resolve any URI that looks like an absolute or
        relative path, or if the URI explicitly has the "file://" header.

        When the "file://" header is used, we expect an absolute path, including
        the leading slash. This means it will look like there are 3 slashes
        after the "file:", for example: "file:///home/me/somefile.txt".

        If this is given a URI with what looks like another URI header (e.g.
        "base64://..."), we thrown an InvalidUriError. This ends up being due to
        the `//` component, which we treat as an invalid path, not because of
        any special parsing.

        :param uri: URI string to resolve into an element instance
        :type uri: str

        :raises smqtk.exceptions.InvalidUriError: This element type could not
            resolve the provided URI string.

        :return: New element instance of our type.
        :rtype: DataElement

        """
        path_match = cls.FILE_URI_RE.match(uri)

        # if did not match RE, then not a valid path to a file (i.e. had a
        # trailing slash, file:// prefix malformed, etc.)
        if path_match is None:
            raise InvalidUriError(uri, "Malformed URI")

        path = path_match.group(1)

        # When given the file:// prefix, the encoded path must be absolute
        # Stealing the notion based on how Google Chrome handles file:// URIs
        if uri.startswith("file://") and not osp.isabs(path):
            raise InvalidUriError(uri, "Found file:// prefix, but path was not "
                                       "absolute")

        return DataFileElement(path)

    def __init__(self, filepath, readonly=False, explicit_mimetype=None):
        """
        Create a new FileElement.

        File the given ``filepath`` refers to may not exist yet.

        :param filepath: Path to the file to wrap.  If relative, it is
            interpreted as relative to the current working directory.
        :type filepath: str

        :param readonly: If this element should allow writing or not.
        :type readonly: bool

        :param explicit_mimetype: Specific mimetype string to use for this
            element. If this is None (default), we try to infer mimetype from
            ``filepath`` extension using the python ``mimetype`` module.
        :type explicit_mimetype: None | str


        Example
        -------
        >>> from smqtk.representation.data_element import file_element
        >>> self = DataFileElement(file_element.__file__)
        >>> assert isinstance(self.get_bytes(), six.binary_type)
        >>> self = DataFileElement('does-not-exist')
        >>> assert isinstance(self.get_bytes(), six.binary_type)
        """
        super(DataFileElement, self).__init__()

        assert isinstance(filepath, six.string_types), \
            "File path must be a string."
        assert isinstance(explicit_mimetype, STR_NONE_TYPES), \
            "Explicit mimetype must either be a string or None."

        # Just expand a user-home `~` if present, keep relative if that's what
        # was given.
        self._filepath = osp.expanduser(filepath)
        self._readonly = bool(readonly)
        self._explicit_mimetype = explicit_mimetype

        self._content_type = explicit_mimetype
        if not self._content_type:
            self._content_type = mimetypes.guess_type(filepath)[0]

    def __repr__(self):
        return super(DataFileElement, self).__repr__() + \
            "{filepath: %s, readonly: %s, explicit_mimetype: %s}" \
            % (self._filepath, self._readonly, self._explicit_mimetype)

    def get_config(self):
        return {
            "filepath": self._filepath,
            "readonly": self._readonly,
            "explicit_mimetype": self._explicit_mimetype,
        }

    #
    # Implemented abstract methods
    #

    def content_type(self):
        """
        :return: Standard type/subtype string for this data element, or None if
            the content type is unknown.
        :rtype: str or None
        """
        return self._content_type

    def is_empty(self):
        """
        Check if this element contains no bytes.

        This plugin checks if the file on disk is greater than 0 in size.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """
        return not osp.exists(self._filepath) or \
            osp.getsize(self._filepath) == 0

    def get_bytes(self):
        """
        :return: Get the byte stream for this data element.
        :rtype: bytes
        """
        return (not self.is_empty() and open(self._filepath, 'rb').read()) or six.b("")

    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        return not self._readonly

    def set_bytes(self, b):
        """
        Set bytes to this data element in the form of a string.

        Not all implementations may support setting bytes (writing). See the
        ``writable`` method.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        if not self._readonly:
            safe_file_write(self._filepath, b)
        else:
            raise ReadOnlyError("This file element is read only.")

    def write_temp(self, temp_dir=None):
        """
        Write this data's bytes to a temporary file on disk, returning the path
        to the written file, whose extension is guessed based on this data's
        content type.

        NOTE:
            The file path returned should not be explicitly removed by the user.
            Instead, the ``clean_temp()`` method should be called on this
            object.

        For FileElement instances, this returns the original data file's path
        unless a `temp_dir` is specified that is not the directory that contains
        the original file.

        :param temp_dir: Optional directory to write temporary file in,
            otherwise we use the platform default temporary files directory.
        :type temp_dir: None or str

        :return: Path to the temporary file
        :rtype: str

        """
        if temp_dir:
            abs_temp_dir = osp.abspath(osp.expanduser(temp_dir))
            # Checking that the dir given isn't where the filepath lives
            if abs_temp_dir != osp.dirname(self._filepath):
                return super(DataFileElement, self).write_temp(temp_dir)
        return self._filepath

    def clean_temp(self):
        """
        Clean any temporary files created by this element. This does nothing if
        no temporary files have been generated for this element.

        For FileElement instance's this does nothing as the ``write_temp()``
        method doesn't actually write any files.
        """
        # does the right thing regardless of what happened in write_temp
        return super(DataFileElement, self).clean_temp()


DATA_ELEMENT_CLASS = DataFileElement



from __future__ import division, print_function
import unittest

import numpy as np

from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.algorithms.relevancy_index.libsvm_hik import LibSvmHikRelevancyIndex


if LibSvmHikRelevancyIndex.is_usable():

    class TestIqrSvmHik (unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            # Don't need to clear cache because we're setting the vectors here
            cls.d0 = DescriptorMemoryElement('index', 0)
            cls.d0.set_vector(np.array([1, 0, 0, 0, 0], float))
            cls.d1 = DescriptorMemoryElement('index', 1)
            cls.d1.set_vector(np.array([0, 1, 0, 0, 0], float))
            cls.d2 = DescriptorMemoryElement('index', 2)
            cls.d2.set_vector(np.array([0, 0, 1, 0, 0], float))
            cls.d3 = DescriptorMemoryElement('index', 3)
            cls.d3.set_vector(np.array([0, 0, 0, 1, 0], float))
            cls.d4 = DescriptorMemoryElement('index', 4)
            cls.d4.set_vector(np.array([0, 0, 0, 0, 1], float))
            cls.d5 = DescriptorMemoryElement('index', 5)
            cls.d5.set_vector(np.array([0.5, 0, 0.5, 0, 0], float))
            cls.d6 = DescriptorMemoryElement('index', 6)
            cls.d6.set_vector(np.array([.2, .2, .2, .2, .2], float))
            cls.index_descriptors = [cls.d0, cls.d1, cls.d2, cls.d3, cls.d4,
                                     cls.d5, cls.d6]

            cls.q_pos = DescriptorMemoryElement('query', 0)
            cls.q_pos.set_vector(np.array([.75, .25, 0, 0, 0], float))
            cls.q_neg = DescriptorMemoryElement('query', 1)
            cls.q_neg.set_vector(np.array([0,   0,   0, .5, .5], float))

        def test_configuration(self):
            c = LibSvmHikRelevancyIndex.get_default_config()
            self.assertIn('descr_cache_filepath', c)

            # change default for something different
            c['descr_cache_filepath'] = 'foobar.thing'

            #: :type: LibSvmHikRelevancyIndex
            iqr_index = LibSvmHikRelevancyIndex.from_config(c)
            self.assertEqual(iqr_index.descr_cache_fp,
                             c['descr_cache_filepath'])

            # test config idempotency
            self.assertDictEqual(c, iqr_index.get_config())

        def test_rank_no_neg(self):
            iqr_index = LibSvmHikRelevancyIndex()
            iqr_index.build_index(self.index_descriptors)
            # index should auto-select some negative examples, thus not raising
            # an exception.
            iqr_index.rank([self.q_pos], [])

        def test_rank_no_pos(self):
            iqr_index = LibSvmHikRelevancyIndex()
            iqr_index.build_index(self.index_descriptors)
            self.assertRaises(ValueError, iqr_index.rank, [], [self.q_neg])

        def test_rank_no_input(self):
            iqr_index = LibSvmHikRelevancyIndex()
            iqr_index.build_index(self.index_descriptors)
            self.assertRaises(ValueError, iqr_index.rank, [], [])

        def test_count(self):
            iqr_index = LibSvmHikRelevancyIndex()
            self.assertEqual(iqr_index.count(), 0)
            iqr_index.build_index(self.index_descriptors)
            self.assertEqual(iqr_index.count(), 7)

        def test_simple_iqr_scenario(self):
            # Make some descriptors;
            # Pick some from created set that are close to each other and use as
            #   positive query, picking some other random descriptors as
            #   negative examples.
            # Rank index based on chosen pos/neg
            # Check that positive choices are at the top of the ranking (closest
            #   to 0) and negative choices are closest to the bottom.
            iqr_index = LibSvmHikRelevancyIndex()
            iqr_index.build_index(self.index_descriptors)

            rank = iqr_index.rank([self.q_pos], [self.q_neg])
            rank_ordered = sorted(rank.items(), key=lambda e: e[1],
                                  reverse=True)

            print("rank_ordered:")
            for i, r in enumerate(rank_ordered):
                print("..{}: {}".format(i, r))

            # Check expected ordering
            # 0-5-1-2-6-3-4
            # - 2 should end up coming before 6, because 6 has more intersection
            #   with the negative example.
            assert rank_ordered[0][0] == self.d0
            assert rank_ordered[1][0] == self.d5
            assert rank_ordered[2][0] == self.d1
            # Results show that d2 and d6 have the same rank, so their position
            # in interchangeable.
            assert rank_ordered[3][0] in (self.d2, self.d6)
            assert rank_ordered[4][0] in (self.d2, self.d6)
            assert rank_ordered[3][0] != rank_ordered[4][0]
            # d3 and d4 evaluate to the same rank based on query (no
            # intersection with positive, equal intersection with negative).
            assert rank_ordered[5][0] in (self.d3, self.d4)
            assert rank_ordered[6][0] in (self.d3, self.d4)
            assert rank_ordered[5][0] != rank_ordered[6][0]



"""
Helper utility for cross validating a supervised classifier configuration.
The classifier used should NOT be configured to save its model since this
process requires us to train the classifier multiple times.

Configuration
-------------
    - plugins
        - supervised_classifier
            Supervised Classifier implementation configuration to use. This
            should not be set to use a persistent model if able (this utility
            will repeatedly train a new model for each fold).

        - descriptor_index
            Index to draw descriptors to classify from.

    - cross_validation
        - truth_labels
            Path to a CSV file containing descriptor UUID the truth label
            associations. This defines what descriptors are used from the given
            index. We error if any descriptor UUIDs listed here are not
            available in the given descriptor index. This file should be in
            [uuid, label] column format.

        - num_folds
            Number of folds to make for cross validation.

        - random_seed
            Optional fixed seed for the

        - classification_use_multiprocessing
            If we should use multiprocessing (vs threading) when classifying
            elements.

    - pr_curves
        - enabled
            If Precision/Recall plots should be generated.

        - show
            If we should attempt to show the graph after it has been generated
            (matplotlib).

        - output_directory
            Directory to save generated plots to. If None, we will not save
            plots. Otherwise we will create the directory (and required parent
            directories) if it does not exist.

        - file_prefix
            String prefix to prepend to standard plot file names.

    - roc_curves
        - enabled
            If ROC curves should be generated

        - show
            If we should attempt to show the plot after it has been generated
            (matplotlib).

        - output_directory
            Directory to save generated plots to. If None, we will not save
            plots. Otherwise we will create the directory (and required parent
            directories) if it does not exist.

        - file_prefix
            String prefix to prepend to standard plot file names.
"""

import csv
import logging
import os

import matplotlib.pyplot as plt
import numpy
import six
import sklearn.model_selection
import sklearn.metrics
import six

from smqtk.algorithms import get_classifier_impls
from smqtk.algorithms.classifier import SupervisedClassifier
from smqtk.representation import (
    ClassificationElementFactory,
    get_descriptor_index_impls,
)
from smqtk.representation.classification_element.memory import \
    MemoryClassificationElement
from smqtk.utils import (
    bin_utils,
    file_utils,
    plugin,
)


__author__ = "paul.tunison@kitware.com"


def get_supervised_classifier_impls():
    return get_classifier_impls(sub_interface=SupervisedClassifier)


def default_config():
    return {
        "plugins": {
            "supervised_classifier":
                plugin.make_config(get_supervised_classifier_impls()),
            "descriptor_index":
                plugin.make_config(get_descriptor_index_impls()),
        },
        "cross_validation": {
            "truth_labels": None,
            "num_folds": 6,
            "random_seed": None,
            "classification_use_multiprocessing": True,
        },
        "pr_curves": {
            "enabled": True,
            "show": False,
            "output_directory": None,
            "file_prefix": None,
        },
        "roc_curves": {
            "enabled": True,
            "show": False,
            "output_directory": None,
            "file_prefix": None,
        },
    }


def cli_parser():
    return bin_utils.basic_cli_parser(__doc__)


def classifier_kfold_validation():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    #
    # Load configurations / Setup data
    #
    use_mp = config['cross_validation']['classification_use_multiprocessing']

    pr_enabled = config['pr_curves']['enabled']
    pr_output_dir = config['pr_curves']['output_directory']
    pr_file_prefix = config['pr_curves']['file_prefix'] or ''
    pr_show = config['pr_curves']['show']

    roc_enabled = config['roc_curves']['enabled']
    roc_output_dir = config['roc_curves']['output_directory']
    roc_file_prefix = config['roc_curves']['file_prefix'] or ''
    roc_show = config['roc_curves']['show']

    log.info("Initializing DescriptorIndex (%s)",
             config['plugins']['descriptor_index']['type'])
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['plugins']['descriptor_index'],
        get_descriptor_index_impls()
    )
    log.info("Loading classifier configuration")
    #: :type: dict
    classifier_config = config['plugins']['supervised_classifier']

    # Always use in-memory ClassificationElement since we are retraining the
    # classifier and don't want possible element caching
    #: :type: ClassificationElementFactory
    classification_factory = ClassificationElementFactory(
        MemoryClassificationElement, {}
    )

    log.info("Loading truth data")
    #: :type: list[str]
    uuids = []
    #: :type: list[str]
    truth_labels = []
    with open(config['cross_validation']['truth_labels']) as f:
        f_csv = csv.reader(f)
        for row in f_csv:
            uuids.append(row[0])
            truth_labels.append(row[1])
    #: :type: numpy.ndarray[str]
    uuids = numpy.array(uuids)
    #: :type: numpy.ndarray[str]
    truth_labels = numpy.array(truth_labels)

    #
    # Cross validation
    #
    kfolds = sklearn.model_selection.StratifiedKFold(
        n_splits=config['cross_validation']['num_folds'],
        shuffle=True,
        random_state=config['cross_validation']['random_seed'],
    ).split(numpy.zeros(len(truth_labels)), truth_labels)

    """
    Truth and classification probability results for test data per fold.
    Format:
        {
            0: {
                '<label>':  {
                    "truth": [...],   # Parallel truth and classification
                    "proba": [...],   # probability values
                },
                ...
            },
            ...
        }
    """
    fold_data = {}

    i = 0
    for train, test in kfolds:
        log.info("Fold %d", i)
        log.info("-- %d training examples", len(train))
        log.info("-- %d test examples", len(test))
        fold_data[i] = {}

        log.info("-- creating classifier")
        #: :type: SupervisedClassifier
        classifier = plugin.from_plugin_config(
            classifier_config,
            get_supervised_classifier_impls()
        )

        log.info("-- gathering descriptors")
        #: :type: dict[str, list[smqtk.representation.DescriptorElement]]
        pos_map = {}
        for idx in train:
            if truth_labels[idx] not in pos_map:
                pos_map[truth_labels[idx]] = []
            pos_map[truth_labels[idx]].append(
                descriptor_index.get_descriptor(uuids[idx])
            )

        log.info("-- Training classifier")
        classifier.train(pos_map)

        log.info("-- Classifying test set")
        m = classifier.classify_async(
            (descriptor_index.get_descriptor(uuids[idx]) for idx in test),
            classification_factory,
            use_multiprocessing=use_mp, ri=1.0
        )
        uuid2c = dict((d.uuid(), c.get_classification())
                      for d, c in six.iteritems(m))

        log.info("-- Pairing truth and computed probabilities")
        # Only considering positive labels
        for t_label in pos_map:
            fold_data[i][t_label] = {
                "truth": [L == t_label for L in truth_labels[test]],
                "proba": [uuid2c[uuid][t_label] for uuid in uuids[test]]
            }

        i += 1

    #
    # Curve generation
    #
    if pr_enabled:
        make_pr_curves(fold_data, pr_output_dir, pr_file_prefix, pr_show)
    if roc_enabled:
        make_roc_curves(fold_data, roc_output_dir, roc_file_prefix, roc_show)


def format_plt(title, x_label, y_label):
    plt.xlim([0., 1.])
    plt.ylim([0., 1.05])
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.legend(loc='best')


def save_plt(output_dir, file_name, show):
    file_utils.safe_create_dir(output_dir)
    save_path = os.path.join(output_dir, file_name)
    plt.savefig(save_path)
    if show:
        plt.show()


def make_curves(log, skl_curve_func, title_hook, x_label, y_label, fold_data,
                output_dir, plot_prefix, show):
    """
    Generic method for PR/ROC curve generation

    fold data format:
        {
            0: {
                '<label>':  {
                    "truth": [...],   # Parallel truth and classification
                    "proba": [...],   # Parallel probability values
                },
                ...
            },
            ...
        }

    :param log: Logger instance to use.
    :param skl_curve_func: scikit-learn curve generation function. This should
        be wrapped to return (x, y) value arrays.
    :param title_hook: Descriptive label of the classifier.
    :param x_label: X-axis label
    :param y_label: Y-axis label
    :param fold_data: Truth and classification probability results for test data
        per fold.  See above for format.
    :param output_dir: Directory to output plot images to.
    :param plot_prefix: String prefix for output files.
    :param show: Show the output plots interactively or not.
    """
    file_utils.safe_create_dir(output_dir)

    log.info("Generating %s curves for per-folds and overall", title_hook)

    # All class labels encountered
    class_labels = set()

    # Make curves for classes per fold. One line per class
    for i in fold_data:
        log.info("-- Fold %i", i)

        plt.clf()
        for label in fold_data[i]:
            log.info("   -- label '%s'", label)
            class_labels.add(label)
            l_truth = fold_data[i][label]['truth']
            l_proba = fold_data[i][label]['proba']
            x, y = skl_curve_func(l_truth, l_proba)
            auc = sklearn.metrics.auc(x, y)
            plt.plot(x, y, label="class '%s' (auc=%f)" % (label, auc))

        format_plt("Classifier %s - Fold %d" % (title_hook, i),
                   x_label, y_label)
        filename = plot_prefix + 'fold_%d.png' % i
        save_plt(output_dir, filename, show)

    # Plot aggregate performance curve per class
    log.info("-- All folds")
    plt.clf()
    for label in sorted(class_labels):
        l_truth = [t for i in fold_data for t in fold_data[i][label]['truth']]
        l_proba = [p for i in fold_data for p in fold_data[i][label]['proba']]
        x, y = skl_curve_func(l_truth, l_proba)
        auc = sklearn.metrics.auc(x, y)
        plt.plot(x, y, label="agg '%s' (auc=%f)" % (label, auc))

    format_plt("Classifier %s - Fold Summary" % title_hook, x_label, y_label)
    filename = plot_prefix + "validation.png"
    save_plt(output_dir, filename, show)


def make_pr_curves(fold_data, output_dir, plot_prefix, show):
    log = logging.getLogger(__name__)

    def skl_pr_curve(truth, proba):
        p, r, _ = sklearn.metrics.precision_recall_curve(truth, proba)
        return r, p

    make_curves(log, skl_pr_curve, "PR", "Recall", "Precision", fold_data,
                output_dir, plot_prefix + 'pr.', show)


def make_roc_curves(fold_data, output_dir, plot_prefix, show):
    log = logging.getLogger(__name__)

    def skl_roc_curve(truth, proba):
        fpr, tpr, _ = sklearn.metrics.roc_curve(truth, proba)
        return fpr, tpr

    make_curves(log, skl_roc_curve, "ROC", "False Positive Rate",
                "True Positive Rate", fold_data, output_dir,
                plot_prefix + 'roc.', show)


if __name__ == '__main__':
    classifier_kfold_validation()



"""
Tests for DataElement abstract interface class methods that provide
functionality.
"""
import hashlib
import mock
import os.path as osp
import tempfile
import unittest
import six

import smqtk.exceptions
import smqtk.representation.data_element

if six.PY2:
    builtin_open = '__builtin__.open'
else:
    builtin_open = 'builtins.open'


# because this has a stable mimetype conversion
EXPECTED_CONTENT_TYPE = "image/png"
EXPECTED_BYTES = six.b("hello world")
EXPECTED_MD5 = hashlib.md5(EXPECTED_BYTES).hexdigest()
EXPECTED_SHA1 = hashlib.sha1(EXPECTED_BYTES).hexdigest()
EXPECTED_SHA512 = hashlib.sha512(EXPECTED_BYTES).hexdigest()
# UUID is currently set to be equivalent to SHA1 value by default
EXPECTED_UUID = EXPECTED_SHA1


# Caches the temp directory before we start mocking things out that would
# otherwise be required for the tempfile module to determine the temp
# directory.
tempfile.gettempdir()


# noinspection PyClassHasNoInit,PyAbstractClass
class DummyDataElement (smqtk.representation.data_element.DataElement):

    TEST_WRITABLE = True
    TEST_BYTES = EXPECTED_BYTES
    TEST_CONTENT_TYPE = EXPECTED_CONTENT_TYPE

    @classmethod
    def is_usable(cls):
        return True

    def __repr__(self):
        return super(DummyDataElement, self).__repr__()

    def get_config(self):
        return {}

    def content_type(self):
        return self.TEST_CONTENT_TYPE

    def is_empty(self):
        pass

    def get_bytes(self):
        return self.TEST_BYTES

    def set_bytes(self, b):
        super(DummyDataElement, self).set_bytes(b)
        self.TEST_BYTES = b

    def writable(self):
        return self.TEST_WRITABLE


class TestDataElementAbstract (unittest.TestCase):

    def test_from_uri_default(self):
        self.assertRaises(
            smqtk.exceptions.NoUriResolutionError,
            DummyDataElement.from_uri, 'some uri'
        )

    def test_not_hashable(self):
        # Hash should be that of the UUID of the element
        de = DummyDataElement()
        self.assertRaises(TypeError, hash, de)

    def test_del(self):
        de = DummyDataElement()
        m_clean_temp = de.clean_temp = mock.Mock()
        del de

        self.assertTrue(m_clean_temp.called)

    def test_equality(self):
        # equal when binary content is the same
        e1 = DummyDataElement()
        e2 = DummyDataElement()

        test_content_1 = 'some similar content'
        e1.TEST_BYTES = e2.TEST_BYTES = test_content_1
        self.assertEqual(e1, e2)

        test_content_2 = 'some other bytes'
        e2.TEST_BYTES = test_content_2
        self.assertNotEqual(e1, e2)

    def test_md5(self):
        de = DummyDataElement()
        md5 = de.md5()
        self.assertEqual(md5, EXPECTED_MD5)

    def test_sha1(self):
        de = DummyDataElement()
        sha1 = de.sha1()
        self.assertEqual(sha1, EXPECTED_SHA1)

    def test_sha512(self):
        de = DummyDataElement()
        sha1 = de.sha512()
        self.assertEqual(sha1, EXPECTED_SHA512)

    @mock.patch('smqtk.representation.data_element.file_utils.'
                'safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_content_type_extension(self,
                                    _mock_open, _mock_os_open, _mock_os_close,
                                    _mock_fcntl, _mock_scd):
        de = DummyDataElement()
        de.content_type = mock.Mock(return_value=None)
        fname = de.write_temp()
        self.assertFalse(fname.endswith('.png'))

        fname = DummyDataElement().write_temp()
        self.assertTrue(fname.endswith('.png'))

    # Cases:
    #   - no existing temps, no specific dir
    #   - no existing temps, given specific dir
    #   - existing temps, no specific dir
    #   - existing temps, given specific dir
    #
    # Mocking open, os.open, os.close and fcntl to actual file interaction
    #   - os.open is used under the hood of tempfile to open a file (which also
    #       creates it on disk).

    @mock.patch('smqtk.representation.data_element.file_utils.safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_writeTemp_noExisting_noDir(self,
                                        mock_open, _mock_os_open,
                                        _mock_os_close, _mock_fcntl, mock_scd):
        # no existing temps, no specific dir
        fp = DummyDataElement().write_temp()

        self.assertFalse(mock_scd.called)
        self.assertTrue(mock_open.called)
        self.assertEqual(osp.dirname(fp), tempfile.gettempdir())

    @mock.patch('smqtk.representation.data_element.file_utils.safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_writeTemp_noExisting_givenDir(self,
                                           mock_open, _mock_os_open,
                                           _mock_os_close, _mock_fcntl,
                                           mock_scd):
        # no existing temps, given specific dir
        target_dir = '/some/dir/somewhere'

        fp = DummyDataElement().write_temp(target_dir)

        mock_scd.assert_called_once_with(target_dir)
        self.assertTrue(mock_open.called)
        self.assertNotEqual(osp.dirname(fp), tempfile.gettempdir())
        self.assertEqual(osp.dirname(fp), target_dir)

    @mock.patch("smqtk.representation.data_element.file_element.osp.isfile")
    @mock.patch('smqtk.representation.data_element.file_utils.safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_writeTemp_hasExisting_noDir(self,
                                         mock_open, _mock_os_open,
                                         _mock_os_close, _mock_fcntl, mock_scd,
                                         mock_isfile):
        # Pretend we have existing temps. Will to "write" a temp file to no
        # specific dir, which should not write anything new and just return the
        # last path in the list.
        prev_0 = '/tmp/file.txt'
        prev_1 = '/tmp/file_two.png'

        de = DummyDataElement()
        de._temp_filepath_stack.append(prev_0)
        de._temp_filepath_stack.append(prev_1)

        # Make sure os.path.isfile returns true so we think things in temp
        # stack exist.
        def osp_isfile_se(path):
            if simulate and path in {prev_0, prev_1}:
                return True
            else:
                return False
        simulate = True
        mock_isfile.side_effect = osp_isfile_se

        fp = de.write_temp()

        self.assertFalse(mock_scd.called)
        self.assertFalse(mock_open.called)
        self.assertEqual(fp, prev_1)

        # _temp_filepath_stack files don't exist, so make sure isfile returns
        # false so clean_temp doesn't try to remove files that don't exist.
        simulate = False

    @mock.patch('smqtk.representation.data_element.file_utils.safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_writeTemp_hasExisting_givenNewDir(self, mock_open, _mock_os_open,
                                               _mock_os_close, _mock_fcntl,
                                               mock_scd):
        # existing temps, given specific dir
        prev_0 = '/tmp/file.txt'
        prev_1 = '/tmp/file_two.png'

        target_dir = '/some/specific/dir'

        de = DummyDataElement()
        de._temp_filepath_stack.append(prev_0)
        de._temp_filepath_stack.append(prev_1)

        fp = de.write_temp(temp_dir=target_dir)

        self.assertTrue(mock_scd.called)
        self.assertTrue(mock_open.called)
        self.assertEqual(osp.dirname(fp), target_dir)

    @mock.patch("smqtk.representation.data_element.file_element.osp.isfile")
    @mock.patch('smqtk.representation.data_element.file_utils.safe_create_dir')
    @mock.patch('fcntl.fcntl')  # global
    @mock.patch('os.close')  # global
    @mock.patch('os.open')  # global
    @mock.patch(builtin_open)
    def test_writeTemp_hasExisting_givenExistingDir(self, mock_open,
                                                    _mock_os_open,
                                                    _mock_os_close,
                                                    _mock_fcntl,
                                                    mock_scd, mock_isfile):
        # Pretend these files already exist as written temp files.
        # We test that write_temp with a target directory yields a previously
        #   "written" temp file.
        #
        # that given specific dir already in stack
        prev_0 = '/dir1/file.txt'
        prev_1 = '/tmp/things/file_two.png'
        prev_2 = '/some/specific/dir'

        def osp_isfile_se(path):
            if simulate and path in {prev_0, prev_1, prev_2}:
                return True
            else:
                return False
        simulate = True
        mock_isfile.side_effect = osp_isfile_se

        de = DummyDataElement()
        de._temp_filepath_stack.append(prev_0)
        de._temp_filepath_stack.append(prev_1)
        de._temp_filepath_stack.append(prev_2)

        target_dir = "/tmp/things"

        # Make sure os.path.isfile returns true so we think things in temp
        # stack exist.
        mock_isfile.return_value = True

        fp = de.write_temp(temp_dir=target_dir)

        self.assertFalse(mock_scd.called)
        self.assertFalse(mock_open.called)
        self.assertEqual(fp, prev_1)

        # _temp_filepath_stack files don't exist, so make sure isfile returns
        # false so clean_temp doesn't try to remove files that don't exist.
        simulate = False

    @mock.patch("smqtk.representation.data_element.os")
    def test_cleanTemp_noTemp(self, mock_os):
        # should do all of nothing
        de = DummyDataElement()

        de.clean_temp()

        self.assertFalse(mock_os.path.isfile.called)
        self.assertFalse(mock_os.remove.called)

    @mock.patch("smqtk.representation.data_element.os")
    def test_cleanTemp_hasTemp_badPath(self, mock_os):
        de = DummyDataElement()
        de._temp_filepath_stack.append('tmp/thing')
        mock_os.path.isfile.return_value = False

        de.clean_temp()

        mock_os.path.isfile.assert_called_once_with('tmp/thing')
        self.assertFalse(mock_os.remove.called)

    @mock.patch("smqtk.representation.data_element.os")
    def test_cleanTemp_hasTemp_validPath(self, mock_os):
        expected_path = '/tmp/something'

        de = DummyDataElement()
        de._temp_filepath_stack.append(expected_path)
        mock_os.path.isfile.return_value = True

        de.clean_temp()

        mock_os.path.isfile.assert_called_once_with(expected_path)
        mock_os.remove.assert_called_once_with(expected_path)

    def test_uuid(self):
        de = DummyDataElement()
        de.TEST_BYTES = EXPECTED_BYTES
        self.assertEqual(de.uuid(), EXPECTED_UUID)

    def test_to_buffered_reader(self):
        # Check that we get expected file-like returns.
        de = DummyDataElement()
        de.TEST_BYTES = EXPECTED_BYTES
        br = de.to_buffered_reader()
        self.assertEqual(br.readlines(), [six.b('hello world')])

        de.TEST_BYTES = six.b('some content\nwith new \nlines')
        br = de.to_buffered_reader()
        self.assertEqual(br.readlines(),
                         [six.b('some content\n'),
                          six.b('with new \n'),
                          six.b('lines')])

    def test_is_read_only(self):
        de = DummyDataElement()
        de.TEST_WRITABLE = True
        self.assertFalse(de.is_read_only())
        de.TEST_WRITABLE = False
        self.assertTrue(de.is_read_only())

    def test_set_bytes_not_writable(self):
        de = DummyDataElement()
        # trigger UUID cache at least once
        self.assertEqual(de.uuid(), EXPECTED_UUID)

        de.TEST_WRITABLE = False
        self.assertRaises(
            smqtk.exceptions.ReadOnlyError,
            de.set_bytes, six.b('test bytes')
        )

        # Caches shouldn't have been invalidated due to error
        self.assertEqual(de.uuid(), EXPECTED_UUID)

    def test_set_bytes_checksum_cache_invalidation(self):
        de = DummyDataElement()
        # trigger UUID cache at least once
        self.assertEqual(de.uuid(), EXPECTED_UUID)

        new_expected_bytes = six.b('some new byte content')
        new_expected_uuid = hashlib.sha1(new_expected_bytes).hexdigest()

        de.TEST_WRITABLE = True
        de.set_bytes(new_expected_bytes)

        # Caches should have been invalidated, so UUID return should now
        # reflect new byte content.
        self.assertNotEqual(de.uuid(), EXPECTED_UUID)
        self.assertEqual(de.uuid(), new_expected_uuid)
